# mautrix-telegram - A Matrix-Telegram puppeting bridge
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, AsyncGenerator, AsyncIterable, Awaitable, NamedTuple, cast
from datetime import datetime
import asyncio
import time

from telethon.errors import (
    AuthKeyDuplicatedError,
    AuthKeyError,
    AuthKeyNotFound,
    RPCError,
    TakeoutInitDelayError,
    UnauthorizedError,
)
from telethon.tl.custom import Dialog
from telethon.tl.functions.account import UpdateStatusRequest
from telethon.tl.functions.contacts import GetContactsRequest, SearchRequest
from telethon.tl.functions.help import GetAppConfigRequest
from telethon.tl.functions.messages import GetAvailableReactionsRequest
from telethon.tl.functions.updates import GetStateRequest
from telethon.tl.functions.users import GetUsersRequest
from telethon.tl.types import (
    Chat,
    ChatForbidden,
    InputUserSelf,
    Message,
    MessageActionContactSignUp,
    MessageActionHistoryClear,
    MessageService,
    NotifyPeer,
    PeerUser,
    TypeUpdate,
    UpdateFolderPeers,
    UpdateNewChannelMessage,
    UpdateNewMessage,
    UpdateNotifySettings,
    UpdatePinnedDialogs,
    UpdateShortChatMessage,
    UpdateShortMessage,
    User as TLUser,
)
from telethon.tl.types.contacts import ContactsNotModified
from telethon.tl.types.help import AppConfig
from telethon.tl.types.messages import AvailableReactions

from mautrix.appservice import DOUBLE_PUPPET_SOURCE_KEY
from mautrix.bridge import BaseUser, async_getter_lock
from mautrix.client import Client
from mautrix.errors import MatrixRequestError, MNotFound
from mautrix.types import PushActionType, PushRuleKind, PushRuleScope, RoomID, RoomTagInfo, UserID
from mautrix.util import background_task
from mautrix.util.bridge_state import BridgeState, BridgeStateEvent
from mautrix.util.opt_prometheus import Gauge

from . import portal as po, puppet as pu, util
from .abstract_user import AbstractUser
from .db import Backfill, BackfillType, Message as DBMessage, PgSession, User as DBUser
from .tgclient import MautrixTelegramClient
from .types import TelegramID

if TYPE_CHECKING:
    from .__main__ import TelegramBridge

SearchResult = NamedTuple("SearchResult", puppet="pu.Puppet", similarity=int)

METRIC_LOGGED_IN = Gauge("bridge_logged_in", "Users logged into bridge")
METRIC_CONNECTED = Gauge("bridge_connected", "Users connected to Telegram")

BridgeState.human_readable_errors.update(
    {
        "tg-not-connected": "Your Telegram connection failed",
        "tg-auth-key-duplicated": "The bridge accidentally logged you out",
        "tg-not-authenticated": "The stored auth token did not work",
        "tg-no-auth": "You're not logged in",
    }
)


class User(DBUser, AbstractUser, BaseUser):
    by_mxid: dict[str, User] = {}
    by_tgid: dict[int, User] = {}

    _portals_cache: dict[tuple[TelegramID, TelegramID], po.Portal] | None

    _ensure_started_lock: asyncio.Lock
    _track_connection_task: asyncio.Task | None
    _backfill_task: asyncio.Task | None
    wakeup_backfill_task: asyncio.Event
    _is_backfilling: bool
    takeout_retry_immediate: asyncio.Event
    takeout_requested: bool

    _available_emoji_reactions: set[str] | None
    _available_emoji_reactions_hash: int | None
    _available_emoji_reactions_fetched: float
    _available_emoji_reactions_lock: asyncio.Lock
    _app_config: dict[str, Any] | None
    _app_config_hash: int

    def __init__(
        self,
        mxid: UserID,
        tgid: TelegramID | None = None,
        tg_username: str | None = None,
        tg_phone: str | None = None,
        is_bot: bool = False,
        is_premium: bool = False,
        saved_contacts: int = 0,
    ) -> None:
        super().__init__(
            mxid=mxid,
            tgid=tgid,
            tg_username=tg_username,
            tg_phone=tg_phone,
            is_bot=is_bot,
            is_premium=is_premium,
            saved_contacts=saved_contacts,
        )
        AbstractUser.__init__(self)
        BaseUser.__init__(self)
        self._ensure_started_lock = asyncio.Lock()
        self._track_connection_task = None
        self._is_backfilling = False
        self._portals_cache = None

        self._backfill_task = None
        self.wakeup_backfill_task = asyncio.Event()
        self.takeout_retry_immediate = asyncio.Event()
        self.takeout_requested = False

        self._available_emoji_reactions = None
        self._available_emoji_reactions_hash = None
        self._available_emoji_reactions_fetched = 0
        self._available_emoji_reactions_lock = asyncio.Lock()
        self._app_config = None
        self._app_config_hash = 0

        (
            self.relaybot_whitelisted,
            self.whitelisted,
            self.puppet_whitelisted,
            self.matrix_puppet_whitelisted,
            self.is_admin,
            self.permissions,
        ) = self.config.get_permissions(self.mxid)

    @property
    def name(self) -> str:
        return self.mxid

    @property
    def mxid_localpart(self) -> str:
        localpart, server = Client.parse_user_id(self.mxid)
        return localpart

    @property
    def human_tg_id(self) -> str:
        return f"@{self.tg_username}" if self.tg_username else f"+{self.tg_phone}" or None

    @property
    def peer(self) -> PeerUser | None:
        return PeerUser(user_id=self.tgid) if self.tgid else None

    # TODO replace with proper displayname getting everywhere
    @property
    def displayname(self) -> str:
        return self.mxid_localpart

    @property
    def plain_displayname(self) -> str:
        return self.displayname

    @classmethod
    def init_cls(cls, bridge: "TelegramBridge") -> AsyncIterable[Awaitable[User]]:
        cls.config = bridge.config
        cls.bridge = bridge
        cls.az = bridge.az
        cls.loop = bridge.loop

        return (user.try_ensure_started() async for user in cls.all_with_tgid())

    # region Telegram connection management

    async def try_ensure_started(self) -> None:
        try:
            await self.ensure_started()
        except Exception:
            self.log.exception("Exception in ensure_started")
        else:
            if not self.client and not await PgSession.has(self.mxid):
                self.log.warning("Didn't start user: no session stored")
                if self.tgid:
                    await self.push_bridge_state(
                        BridgeStateEvent.BAD_CREDENTIALS, error="tg-no-auth"
                    )

    async def ensure_started(self, even_if_no_session=False) -> User:
        if not self.puppet_whitelisted or self.connected:
            return self
        async with self._ensure_started_lock:
            return cast(User, await super().ensure_started(even_if_no_session))

    async def on_signed_out(self, err: UnauthorizedError | AuthKeyError | AuthKeyNotFound) -> None:
        error_code = "tg-auth-error"
        if isinstance(err, AuthKeyDuplicatedError):
            error_code = "tg-auth-key-duplicated"
            message = None
        else:
            message = str(err)
        self.log.warning(f"User got signed out with {err}, deleting data...")
        try:
            await self.log_out(
                state=BridgeStateEvent.BAD_CREDENTIALS,
                error=error_code,
                message=message,
                delete=False,
            )
        except Exception:
            self.log.exception("Error handling external logout")

    async def start(self, delete_unless_authenticated: bool = False) -> User:
        try:
            await super().start()
        except (AuthKeyDuplicatedError, AuthKeyNotFound) as e:
            self.log.warning(f"Got {type(e).__name__} in start()")
            await self.on_signed_out(e)
            if not delete_unless_authenticated:
                # The caller wants the client to be connected, so restart the connection.
                await super().start()
            return self
        except Exception:
            await self.push_bridge_state(BridgeStateEvent.UNKNOWN_ERROR)
            raise
        try:
            assert self.client, "client is undefined"
            assert self.client.is_connected(), "client is not connected"
            await self.client(GetStateRequest())
        except AssertionError as e:
            self.log.error(f"Client in bad state after start(): {e}")
            if self.tgid:
                await self.push_bridge_state(BridgeStateEvent.UNKNOWN_ERROR, message=str(e))
        except UnauthorizedError as e:
            if delete_unless_authenticated or self.tgid:
                self.log.error(f"Authorization error in start(): {type(e)}: {e}")
            if self.tgid:
                await self.on_signed_out(e)
        except RPCError as e:
            self.log.error(f"Unknown RPC error in start(): {type(e)}: {e}")
            if self.tgid:
                await self.push_bridge_state(BridgeStateEvent.UNKNOWN_ERROR, message=str(e))
        else:
            # Authenticated, run post login
            self.log.debug(f"Ensuring post_login() for {self.name}")
            background_task.create(self.post_login())
            return self
        # Not authenticated, delete data if necessary
        if delete_unless_authenticated and self.client is not None:
            self.log.debug(f"Unauthenticated user {self.name} start()ed, deleting session...")
            await self.client.disconnect()
            await self.client.session.delete()
        return self

    @property
    def _is_connected(self) -> bool:
        return bool(
            self.client and self.client._sender and self.client._sender._transport_connected()
        )

    @property
    def _bridge_state_info(self) -> dict[str, Any]:
        if self.takeout_requested:
            return {
                "takeout_requested": True,
            }
        return {}

    async def _track_connection(self) -> None:
        self.log.debug("Starting loop to track connection state")
        while True:
            await asyncio.sleep(3)
            connected = self._is_connected
            self._track_metric(METRIC_CONNECTED, connected)
            if connected:
                await self.push_bridge_state(
                    BridgeStateEvent.BACKFILLING
                    if self._is_backfilling
                    else BridgeStateEvent.CONNECTED,
                    info=self._bridge_state_info,
                )
            else:
                await self.push_bridge_state(
                    BridgeStateEvent.TRANSIENT_DISCONNECT, error="tg-not-connected"
                )

    async def fill_bridge_state(self, state: BridgeState) -> None:
        await super().fill_bridge_state(state)
        if self.tgid:
            state.remote_id = str(self.tgid)
            state.remote_name = self.human_tg_id

    async def get_bridge_states(self) -> list[BridgeState]:
        if not self.tgid:
            return []
        if self._is_connected and await self.is_logged_in():
            state_event = (
                BridgeStateEvent.BACKFILLING
                if self._is_backfilling
                else BridgeStateEvent.CONNECTED
            )
            ttl = 3600
        else:
            state_event = BridgeStateEvent.UNKNOWN_ERROR
            ttl = 240
        return [BridgeState(state_event=state_event, ttl=ttl, info=self._bridge_state_info)]

    async def get_puppet(self) -> pu.Puppet | None:
        if not self.tgid:
            return None
        return await pu.Puppet.get_by_tgid(self.tgid)

    async def get_portal_with(self, puppet: pu.Puppet, create: bool = True) -> po.Portal | None:
        if not self.tgid:
            return None
        return await po.Portal.get_by_tgid(
            puppet.tgid, tg_receiver=self.tgid, peer_type="user" if create else None
        )

    async def stop(self) -> None:
        if self._track_connection_task:
            self._track_connection_task.cancel()
            self._track_connection_task = None
        if self._backfill_task:
            self._backfill_task.cancel()
            self._backfill_task = None
        await super().stop()
        self._track_metric(METRIC_CONNECTED, False)

    async def post_login(self, info: TLUser = None, first_login: bool = False) -> None:
        if (
            self.config["metrics.enabled"] or self.config["homeserver.status_endpoint"]
        ) and not self._track_connection_task:
            self._track_connection_task = asyncio.create_task(self._track_connection())

        try:
            await self.update_info(info)
        except Exception:
            self.log.exception("Failed to update telegram account info")
            return

        self._track_metric(METRIC_LOGGED_IN, True)
        if not self._backfill_task or self._backfill_task.done():
            self._backfill_task = asyncio.create_task(self._try_handle_backfill_requests_loop())

        try:
            puppet = await pu.Puppet.get_by_tgid(self.tgid)
            if puppet.custom_mxid != self.mxid and puppet.can_auto_login(self.mxid):
                self.log.info(f"Automatically enabling custom puppet")
                await puppet.switch_mxid(access_token="auto", mxid=self.mxid)
        except Exception:
            self.log.exception("Failed to automatically enable custom puppet")

        if not self.is_bot and (self.config["bridge.startup_sync"] or first_login):
            try:
                self._is_backfilling = True
                await self.sync_dialogs()
                await self.sync_contacts()
            except Exception:
                self.log.exception("Failed to run post-login sync")
            finally:
                self._is_backfilling = False

    @property
    def _takeout_options(self) -> dict[str, bool | int]:
        return {
            "users": True,
            "chats": self.config["bridge.backfill.normal_groups"],
            "megagroups": True,
            "channels": True,
            "files": True,
            "max_file_size": min(self.bridge.matrix.media_config.upload_size, 2000 * 1024 * 1024),
        }

    async def _try_handle_backfill_requests_loop(self) -> None:
        if not self.config["bridge.backfill.enable"]:
            return
        try:
            await self._handle_backfill_requests_loop()
        except Exception:
            self.log.exception("Fatal error in backfill request loop")

    async def _handle_backfill_requests_loop(self) -> None:
        while True:
            req = await Backfill.get_next(self.mxid)
            if not req:
                try:
                    await asyncio.wait_for(self.wakeup_backfill_task.wait(), timeout=300)
                except asyncio.TimeoutError:
                    pass
                self.wakeup_backfill_task.clear()
            else:
                try:
                    await self._takeout_and_backfill(req)
                except Exception:
                    self.log.exception("Error in takeout backfill loop, retrying in an hour")
                    await asyncio.sleep(3600)

    async def _check_server_notice_edit(self, message: Message) -> None:
        if "Data export request" in message.message and "Accepted" in message.message:
            self.log.debug(
                f"Received an edit to message {message.id} that looks like the data export"
                " was accepted, marking takeout as retriable"
            )
            self.takeout_retry_immediate.set()

    async def _takeout_and_backfill(self, first_req: Backfill, first_attempt: bool = True) -> None:
        self.takeout_retry_immediate.clear()
        self.takeout_requested = True
        try:
            async with self.client.takeout(**self._takeout_options) as takeout_client:
                self.takeout_requested = False
                self.log.info("Acquired takeout client successfully")
                await self._backfill_loop_with_client(takeout_client, first_req)
                self.log.info("Backfills finished, exiting takeout")
        except TakeoutInitDelayError as e:
            if first_attempt:
                self.log.info(
                    f"Takeout requested, will wait for retry request or {e.seconds} seconds"
                )
            else:
                self.log.warning(
                    f"Got takeout init delay again after retry, waiting for {e.seconds} seconds"
                )
            try:
                await asyncio.wait_for(self.takeout_retry_immediate.wait(), timeout=e.seconds)
                self.log.info("Retrying takeout")
            except asyncio.TimeoutError:
                self.log.info("Takeout timeout expired")
            await self._takeout_and_backfill(first_req, first_attempt=False)

    async def _backfill_loop_with_client(
        self, client: MautrixTelegramClient, first_req: Backfill
    ) -> None:
        missed_reqs = 0
        while missed_reqs < 10:
            req = first_req or await Backfill.get_next(self.mxid)
            first_req = None
            if not req:
                missed_reqs += 1
                try:
                    await asyncio.wait_for(self.wakeup_backfill_task.wait(), timeout=30)
                except asyncio.TimeoutError:
                    pass
                self.wakeup_backfill_task.clear()
                continue
            missed_reqs = 0
            self.log.info("Backfill request %s", req)
            try:
                portal = await po.Portal.get_by_tgid(
                    TelegramID(req.portal_tgid), tg_receiver=TelegramID(req.portal_tg_receiver)
                )
                await req.mark_dispatched()
                if req.type == BackfillType.HISTORICAL:
                    await portal.backfill(self, client, req=req)
                elif req.type == BackfillType.SYNC_DIALOG:
                    await self._backfill_sync_dialog(portal, client, req.extra_data)
                await req.mark_done()
                await asyncio.sleep(req.post_batch_delay)
            except Exception:
                self.log.exception("Error handling backfill request for %s", req.portal_tgid)
                await req.set_cooldown_timeout(1800)

    async def _backfill_sync_dialog(
        self, portal: po.Portal, client: MautrixTelegramClient, post_sync_args: dict[str, Any]
    ) -> None:
        if portal.mxid:
            self.log.debug("Portal already exists, skipping dialog sync backfill queue item")
            return
        self.log.info(f"Creating portal for {portal.tgid_log} as part of backfill loop")
        try:
            await portal.create_matrix_room(
                self,
                client=client,
                update_if_exists=False,
                invites=[self.mxid],
                from_dialog_sync=True,
            )
        except Exception:
            self.log.exception(f"Error while creating {portal.tgid_log}")
        else:
            await self.post_sync_dialog(portal, puppet=None, was_created=True, **post_sync_args)

    async def update(self, update: TypeUpdate) -> bool:
        if not self.is_bot:
            return False

        if isinstance(update, (UpdateNewMessage, UpdateNewChannelMessage)):
            portal = await po.Portal.get_by_entity(update.message.peer_id, tg_receiver=self.tgid)
        elif isinstance(update, UpdateShortChatMessage):
            portal = await po.Portal.get_by_tgid(TelegramID(update.chat_id))
        elif isinstance(update, UpdateShortMessage):
            portal = await po.Portal.get_by_tgid(
                TelegramID(update.user_id), tg_receiver=self.tgid, peer_type="user"
            )
        else:
            return False

        if portal:
            await self.register_portal(portal)
            return False

        # Don't bother handling the update
        return True

    # endregion
    # region Telegram actions that need custom methods

    async def set_presence(self, online: bool = True) -> None:
        if not self.is_bot:
            await self.client(UpdateStatusRequest(offline=not online))

    async def get_me(self) -> TLUser | None:
        try:
            return (await self.client(GetUsersRequest([InputUserSelf()])))[0]
        except UnauthorizedError as e:
            self.log.error(f"Authorization error in get_me(): {type(e)}: {e}")
            await self.push_bridge_state(
                BridgeStateEvent.BAD_CREDENTIALS, error="tg-auth-error", message=str(e), ttl=3600
            )
            await self.stop()
            return None

    async def update_info(self, info: TLUser | None = None) -> None:
        if not info:
            info = await self.get_me()
            if not info:
                self.log.warning("get_me() returned None, aborting update_info()")
                return
        changed = False
        if self.is_bot != info.bot:
            self.is_bot = info.bot
            changed = True
        if self.is_premium != info.premium:
            self.is_premium = info.premium
            changed = True
        if self.tg_username != info.username:
            self.tg_username = info.username
            changed = True
        if self.tg_phone != info.phone:
            self.tg_phone = info.phone
            changed = True
        if self.tgid != info.id:
            self.tgid = TelegramID(info.id)
            self.by_tgid[self.tgid] = self
        if changed:
            await self.save()

    async def kick_from_portals(self) -> None:
        if not self.config["bridge.kick_on_logout"]:
            return
        portals = await self.get_cached_portals()
        for portal in portals.values():
            if not portal or portal.deleted or not portal.mxid or portal.has_bot:
                continue
            if portal.peer_type == "user":
                await portal.cleanup_portal("Logged out of Telegram")
            else:
                try:
                    await portal.main_intent.kick_user(
                        portal.mxid, self.mxid, "Logged out of Telegram."
                    )
                except MatrixRequestError:
                    pass

    async def log_out(
        self,
        delete: bool = True,
        do_logout: bool = True,
        state: BridgeStateEvent = BridgeStateEvent.LOGGED_OUT,
        error: str | None = None,
        message: str | None = None,
    ) -> bool:
        puppet = await pu.Puppet.get_by_tgid(self.tgid)
        if puppet is not None and puppet.is_real_user:
            await puppet.switch_mxid(None, None)
        try:
            await self.kick_from_portals()
        except Exception:
            self.log.exception("Failed to kick user from portals on logout")
        if self.tgid:
            try:
                del self.by_tgid[self.tgid]
            except KeyError:
                pass
        ok = False
        if self.client is not None:
            sess = self.client.session
            # Try to send a logout request. If it succeeds, this also disconnects the client and
            # deletes the session, but we do those again later just to be safe.
            if do_logout:
                ok = await self.client.log_out()
            # Force-disconnect the client and set it to None
            await self.stop()
            await sess.delete()

        # Drop LOGGED_OUT states if the user was already logged out previously
        # and doesn't have a remote ID anymore
        # TODO send a management room notice for non-manual logouts?
        if self.tgid or state != BridgeStateEvent.LOGGED_OUT:
            await self.push_bridge_state(state, error=error, message=message)
        if delete:
            await self.delete()
            self.by_mxid.pop(self.mxid, None)
            self.log.info("User deleted")
        else:
            await self.remove_tgid()
            self.log.info("User telegram ID cleared")
        self._track_metric(METRIC_LOGGED_IN, False)
        return ok

    async def _search_local(
        self, query: str, max_results: int = 5, min_similarity: int = 45
    ) -> list[SearchResult]:
        results: list[SearchResult] = []
        for contact_id in await self.get_contacts():
            contact = await pu.Puppet.get_by_tgid(contact_id, create=False)
            if not contact:
                continue
            similarity = contact.similarity(query)
            if similarity >= min_similarity:
                results.append(SearchResult(contact, similarity))
        results.sort(key=lambda tup: tup[1], reverse=True)
        return results[0:max_results]

    async def _search_remote(self, query: str, max_results: int = 5) -> list[SearchResult]:
        if len(query) < 5:
            return []
        server_results = await self.client(SearchRequest(q=query, limit=max_results))
        results: list[SearchResult] = []
        for user in server_results.users:
            puppet = await pu.Puppet.get_by_tgid(user.id)
            await puppet.update_info(self, user)
            results.append(SearchResult(puppet, puppet.similarity(query)))
        results.sort(key=lambda tup: tup[1], reverse=True)
        return results[0:max_results]

    async def search(
        self, query: str, force_remote: bool = False
    ) -> tuple[list[SearchResult], bool]:
        if force_remote:
            return await self._search_remote(query), True

        results = await self._search_local(query)
        if results:
            return results, False

        return await self._search_remote(query), True

    async def get_direct_chats(self) -> dict[UserID, list[RoomID]]:
        return {
            pu.Puppet.get_mxid_from_id(portal.tgid): [portal.mxid]
            async for portal in po.Portal.find_private_chats_of(self.tgid)
            if portal.mxid
        }

    async def _tag_room(
        self, puppet: pu.Puppet, portal: po.Portal, tag: str, active: bool
    ) -> None:
        if not tag or not portal or not portal.mxid:
            return
        tag_info = await puppet.intent.get_room_tag(portal.mxid, tag)
        if active and tag_info is None:
            tag_info = RoomTagInfo(order=0.5)
            tag_info[DOUBLE_PUPPET_SOURCE_KEY] = self.bridge.name
            self.log.debug(f"Adding tag {tag} to {portal.mxid}/{portal.tgid}")
            await puppet.intent.set_room_tag(portal.mxid, tag, tag_info)
        elif (
            not active and tag_info and tag_info.get(DOUBLE_PUPPET_SOURCE_KEY) == self.bridge.name
        ):
            self.log.debug(f"Removing tag {tag} from {portal.mxid}/{portal.tgid}")
            await puppet.intent.remove_room_tag(portal.mxid, tag)

    async def _mute_room(self, puppet: pu.Puppet, portal: po.Portal, mute_until: float) -> None:
        if not self.config["bridge.mute_bridging"] or not portal or not portal.mxid:
            return
        if mute_until is not None and mute_until > time.time():
            self.log.debug(
                f"Muting {portal.mxid}/{portal.tgid} (muted until {mute_until} on Telegram)"
            )
            await puppet.intent.set_push_rule(
                PushRuleScope.GLOBAL,
                PushRuleKind.ROOM,
                portal.mxid,
                actions=[PushActionType.DONT_NOTIFY],
            )
        else:
            try:
                await puppet.intent.remove_push_rule(
                    PushRuleScope.GLOBAL, PushRuleKind.ROOM, portal.mxid
                )
                self.log.debug(f"Unmuted {portal.mxid}/{portal.tgid}")
            except MNotFound:
                pass

    async def update_folder_peers(self, update: UpdateFolderPeers) -> None:
        if self.config["bridge.tag_only_on_create"]:
            return
        puppet = await pu.Puppet.get_by_custom_mxid(self.mxid)
        if not puppet or not puppet.is_real_user:
            return
        for peer in update.folder_peers:
            portal = await po.Portal.get_by_entity(peer.peer, tg_receiver=self.tgid, create=False)
            await self._tag_room(
                puppet, portal, self.config["bridge.archive_tag"], peer.folder_id == 1
            )

    async def update_pinned_dialogs(self, update: UpdatePinnedDialogs) -> None:
        if self.config["bridge.tag_only_on_create"]:
            return
        puppet = await pu.Puppet.get_by_custom_mxid(self.mxid)
        if not puppet or not puppet.is_real_user:
            return
        # TODO bridge unpinning properly
        for pinned in update.order:
            portal = await po.Portal.get_by_entity(
                pinned.peer, tg_receiver=self.tgid, create=False
            )
            await self._tag_room(puppet, portal, self.config["bridge.pinned_tag"], True)

    async def update_notify_settings(self, update: UpdateNotifySettings) -> None:
        if self.config["bridge.tag_only_on_create"]:
            return
        elif not isinstance(update.peer, NotifyPeer):
            # TODO handle global notification setting changes?
            return
        puppet = await pu.Puppet.get_by_custom_mxid(self.mxid)
        if not puppet or not puppet.is_real_user:
            return
        portal = await po.Portal.get_by_entity(
            update.peer.peer, tg_receiver=self.tgid, create=False
        )
        await self._mute_room(puppet, portal, update.notify_settings.mute_until.timestamp())

    @staticmethod
    def dialog_to_sync_args(dialog: Dialog) -> dict:
        return {
            "last_message_ts": (
                cast(datetime, dialog.date).timestamp() if dialog.date else time.time()
            ),
            "unread_count": dialog.unread_count,
            "max_read_id": dialog.dialog.read_inbox_max_id,
            "mute_until": (
                dialog.dialog.notify_settings.mute_until.timestamp()
                if dialog.dialog.notify_settings.mute_until
                else None
            ),
            "pinned": dialog.pinned,
            "archived": dialog.archived,
        }

    async def _sync_dialog(
        self, portal: po.Portal, dialog: Dialog, should_create: bool, puppet: pu.Puppet | None
    ) -> None:
        if (
            not portal.mxid
            and isinstance(dialog.message, MessageService)
            and isinstance(
                dialog.message.action, (MessageActionContactSignUp, MessageActionHistoryClear)
            )
        ):
            self.log.debug(
                f"Not syncing {portal.tgid_log} "
                f"(last message is a {type(dialog.message.action).__name__})"
            )
            return
        was_created = False
        post_sync_args = self.dialog_to_sync_args(dialog)
        if portal.mxid:
            self.log.debug(f"Backfilling and updating {portal.tgid_log} (dialog sync)")
            try:
                await portal.forward_backfill(self, initial=False, last_tgid=dialog.message.id)
            except Exception:
                self.log.exception(f"Error while backfilling {portal.tgid_log}")
            try:
                await portal.update_matrix_room(self, dialog.entity)
            except Exception:
                self.log.exception(f"Error while updating {portal.tgid_log}")
        elif should_create:
            self.log.debug(f"Creating portal for {portal.tgid_log} immediately (dialog sync)")
            try:
                await portal.create_matrix_room(
                    self, dialog.entity, invites=[self.mxid], from_dialog_sync=True
                )
                was_created = True
            except Exception:
                self.log.exception(f"Error while creating {portal.tgid_log}")
        elif self.config["bridge.sync_deferred_create_all"]:
            self.log.debug(f"Enqueuing deferred dialog sync for {portal.tgid_log}")
            await portal.enqueue_backfill(
                self,
                priority=40,
                type=BackfillType.SYNC_DIALOG,
                extra_data=post_sync_args,
            )
        if portal.mxid and puppet and puppet.is_real_user:
            await self.post_sync_dialog(
                portal=portal,
                puppet=puppet,
                was_created=was_created,
                **post_sync_args,
            )
        self.log.debug(f"_sync_dialog finished for {portal.tgid_log}")

    async def post_sync_dialog(
        self,
        portal: po.Portal,
        puppet: pu.Puppet | None,
        was_created: bool,
        max_read_id: int,
        last_message_ts: float,
        unread_count: int,
        mute_until: float,
        pinned: bool,
        archived: bool,
    ) -> None:
        if puppet is None:
            puppet = await pu.Puppet.get_by_custom_mxid(self.mxid)
            if not puppet or not puppet.is_real_user:
                return
        self.log.debug(
            f"Running dialog post-sync for {portal.tgid_log} with args "
            f"{was_created=}, {max_read_id=}, {last_message_ts=}, {unread_count=}, "
            f"{mute_until=}, {pinned=}, {archived=}"
        )
        tg_space = portal.tgid if portal.peer_type == "channel" else self.tgid
        unread_threshold_hours = self.config["bridge.backfill.unread_hours_threshold"]
        force_read = (
            was_created
            and unread_threshold_hours >= 0
            and last_message_ts + (unread_threshold_hours * 60 * 60) < time.time()
        )
        if unread_count == 0 or force_read:
            # This is usually more reliable than finding a specific message
            # e.g. if the last read message is a service message that isn't in the message db
            last_read = await DBMessage.find_last(portal.mxid, tg_space)
            if force_read:
                self.log.debug(
                    f"Marking {portal.tgid_log} as read because the last message is from "
                    f"{last_message_ts} (unread threshold is {unread_threshold_hours} hours)"
                )
        else:
            last_read = await DBMessage.get_one_by_tgid(portal.tgid, tg_space, max_read_id)
        try:
            if last_read:
                await puppet.intent.mark_read(last_read.mx_room, last_read.mxid)
            if was_created or not self.config["bridge.tag_only_on_create"]:
                await self._mute_room(puppet, portal, mute_until)
                await self._tag_room(puppet, portal, self.config["bridge.pinned_tag"], pinned)
                await self._tag_room(puppet, portal, self.config["bridge.archive_tag"], archived)
        except Exception:
            self.log.exception(f"Error updating read status and tags for {portal.tgid_log}")

    async def get_cached_portals(self) -> dict[tuple[TelegramID, TelegramID], po.Portal]:
        if self._portals_cache is None:
            self._portals_cache = {
                (tgid, tg_receiver): await po.Portal.get_by_tgid(tgid, tg_receiver=tg_receiver)
                for tgid, tg_receiver in await self.get_portals()
            }
        return self._portals_cache

    async def sync_dialogs(self) -> None:
        if self.is_bot:
            return
        creators = []
        update_limit = self.config["bridge.sync_update_limit"] or None
        create_limit = self.config["bridge.sync_create_limit"]
        index = 0
        self.log.debug(f"Syncing dialogs ({update_limit=}, {create_limit=})")
        await self.push_bridge_state(BridgeStateEvent.BACKFILLING)
        puppet = await pu.Puppet.get_by_custom_mxid(self.mxid)
        dialog: Dialog
        old_portal_cache = await self.get_cached_portals()
        new_portal_cache = old_portal_cache.copy()
        async for dialog in self.client.iter_dialogs(
            limit=update_limit, ignore_migrated=True, archived=False
        ):
            entity = dialog.entity
            if isinstance(entity, ChatForbidden):
                self.log.warning(f"Ignoring forbidden chat {entity} while syncing")
                continue
            elif isinstance(entity, Chat) and (entity.deactivated or entity.left):
                self.log.warning(f"Ignoring deactivated or left chat {entity} while syncing")
                continue
            elif isinstance(entity, TLUser) and not self.config["bridge.sync_direct_chats"]:
                self.log.trace(f"Ignoring user {entity.id} while syncing")
                continue
            portal = await po.Portal.get_by_entity(entity, tg_receiver=self.tgid)
            new_portal_cache[portal.tgid_full] = portal
            should_create = not create_limit or index < create_limit
            coro = self._sync_dialog(
                portal=portal,
                dialog=dialog,
                puppet=puppet,
                should_create=should_create,
            )
            creators.append(asyncio.create_task(coro))
            index += 1
        if new_portal_cache.keys() != old_portal_cache.keys():
            await self.set_portals(new_portal_cache.keys())
            self._portals_cache = new_portal_cache
        await asyncio.gather(*creators)
        await self.update_direct_chats()
        self.log.debug("Dialog syncing complete")

    async def register_portal(self, portal: po.Portal) -> None:
        self.log.trace(f"Registering portal {portal.tgid_full}")
        if self._portals_cache is not None:
            if self._portals_cache.get(portal.tgid_full) == portal:
                return
            self._portals_cache[portal.tgid_full] = portal
        await super().register_portal(portal.tgid, portal.tg_receiver)

    async def unregister_portal(self, tgid: TelegramID, tg_receiver: TelegramID) -> None:
        self.log.trace(f"Unregistering portal {(tgid, tg_receiver)}")
        if self._portals_cache is not None:
            self._portals_cache.pop((tgid, tg_receiver), None)
        await super().unregister_portal(tgid, tg_receiver)

    async def needs_relaybot(self, portal: po.Portal) -> bool:
        return not await self.is_logged_in() or (
            (portal.has_bot or self.is_bot)
            and portal.tgid_full not in await self.get_cached_portals()
        )

    @staticmethod
    def _hash_contacts(count: int, ids: list[TelegramID]) -> int:
        acc = 0
        for contact in sorted([count] + ids):
            acc = (acc * 20261 + contact) & 0xFFFFFFFF
        return acc & 0x7FFFFFFF

    async def sync_contacts(self, get_info: bool = False) -> dict[TelegramID, dict]:
        existing_contacts = await self.get_contacts()
        contact_hash = self._hash_contacts(self.saved_contacts, existing_contacts)
        response = await self.client(GetContactsRequest(hash=contact_hash))
        if isinstance(response, ContactsNotModified):
            if get_info:
                return {
                    tgid: (await pu.Puppet.get_by_tgid(tgid)).contact_info
                    for tgid in existing_contacts
                }
            return {}
        self.log.debug(f"Updating contacts of {self.name}...")
        if self.saved_contacts != response.saved_count:
            self.saved_contacts = response.saved_count
            await self.save()
        contacts = {}
        for user in response.users:
            puppet: pu.Puppet = await pu.Puppet.get_by_tgid(user.id)
            await puppet.update_info(self, user)
            contacts[user.id] = puppet.contact_info
        await self.set_contacts(contacts.keys())
        self.log.debug("Contact syncing complete")
        return contacts

    @property
    def _available_reactions_up_to_date(self) -> bool:
        return (
            bool(self._available_emoji_reactions)
            and self._available_emoji_reactions_fetched + 12 * 60 * 60 > time.monotonic()
        )

    async def get_available_reactions(self) -> set[str]:
        if self._available_reactions_up_to_date:
            return self._available_emoji_reactions
        async with self._available_emoji_reactions_lock:
            if self._available_reactions_up_to_date:
                return self._available_emoji_reactions
            self.log.debug("Fetching available emoji reactions")
            available_reactions = await self.client(
                GetAvailableReactionsRequest(hash=self._available_emoji_reactions_hash or 0)
            )
            if isinstance(available_reactions, AvailableReactions):
                self._available_emoji_reactions = {
                    react.reaction
                    for react in available_reactions.reactions
                    if not react.inactive and (self.is_premium or not react.premium)
                }
                self._available_emoji_reactions_hash = available_reactions.hash
                self._available_emoji_reactions_fetched = time.monotonic()
                self.log.debug(
                    "Got available emoji reactions: %s", self._available_emoji_reactions
                )
            elif self._available_emoji_reactions is None:
                self.log.warning(
                    f"Got {available_reactions} in response to available reactions request"
                    " even though nothing is cached"
                )
            return self._available_emoji_reactions

    def tl_to_json(self) -> Any:
        pass

    async def get_app_config(self) -> dict[str, Any]:
        if not self._app_config:
            cfg: AppConfig = await self.client(GetAppConfigRequest(hash=self._app_config_hash))
            self._app_config = util.parse_tl_json(cfg.config)
            self._app_config_hash = cfg.hash
        return self._app_config

    async def get_max_reactions(self, is_premium: bool | None = None) -> int:
        if is_premium is None:
            is_premium = self.is_premium
        cfg = await self.get_app_config()
        return (
            cfg.get("reactions_user_max_premium", 3)
            if is_premium
            else cfg.get("reactions_user_max_default", 1)
        )

    # endregion
    # region Class instance lookup

    def _add_to_cache(self) -> None:
        self.by_mxid[self.mxid] = self
        if self.tgid:
            self.by_tgid[self.tgid] = self

    @classmethod
    async def get_and_start_by_mxid(cls, mxid: UserID, even_if_no_session: bool = False) -> User:
        user = await cls.get_by_mxid(mxid, create=True)
        await user.ensure_started(even_if_no_session=even_if_no_session)
        return user

    @classmethod
    async def all_with_tgid(cls) -> AsyncGenerator[User, None]:
        users = await super().all_with_tgid()
        user: cls
        for user in users:
            try:
                yield cls.by_mxid[user.mxid]
            except KeyError:
                user._add_to_cache()
                yield user

    @classmethod
    @async_getter_lock
    async def get_by_mxid(
        cls, mxid: UserID, /, *, check_db: bool = True, create: bool = True
    ) -> User | None:
        if not mxid or pu.Puppet.get_id_from_mxid(mxid) or mxid == cls.az.bot_mxid:
            return None
        try:
            return cls.by_mxid[mxid]
        except KeyError:
            pass

        if not check_db:
            return None

        user = cast(cls, await super().get_by_mxid(mxid))
        if user is not None:
            user._add_to_cache()
            return user

        if create:
            cls.log.debug(f"Creating user instance for {mxid}")
            user = cls(mxid)
            await user.insert()
            user._add_to_cache()
            return user

        return None

    @classmethod
    @async_getter_lock
    async def get_by_tgid(cls, tgid: TelegramID, /) -> User | None:
        try:
            return cls.by_tgid[tgid]
        except KeyError:
            pass

        user = cast(cls, await super().get_by_tgid(tgid))
        if user is not None:
            user._add_to_cache()
            return user

        return None

    @classmethod
    async def find_by_username(cls, username: str) -> User | None:
        if not username:
            return None

        username = username.lower()

        for _, user in cls.by_tgid.items():
            if user.tg_username and user.tg_username.lower() == username:
                return user

        user = cast(cls, await super().find_by_username(username))
        if user:
            try:
                return cls.by_mxid[user.mxid]
            except KeyError:
                user._add_to_cache()
                return user

        return None

    # endregion



# mautrix-telegram - A Matrix-Telegram puppeting bridge
# Copyright (C) 2022 Tulir Asokan, Sumner Evans
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from mautrix.util.async_db import Connection, Scheme

from . import upgrade_table


@upgrade_table.register(description="Add the backfill queue table")
async def upgrade_v11(conn: Connection, scheme: Scheme) -> None:
    gen = ""
    if scheme in (Scheme.POSTGRES, Scheme.COCKROACH):
        gen = "GENERATED ALWAYS AS IDENTITY"
    await conn.execute(
        f"""
        CREATE TABLE backfill_queue (
            queue_id            INTEGER PRIMARY KEY {gen},
            user_mxid           TEXT,
            priority            INTEGER NOT NULL,
            portal_tgid         BIGINT,
            portal_tg_receiver  BIGINT,
            messages_per_batch  INTEGER NOT NULL,
            post_batch_delay    INTEGER NOT NULL,
            max_batches         INTEGER NOT NULL,
            dispatch_time       TIMESTAMP,
            completed_at        TIMESTAMP,
            cooldown_timeout    TIMESTAMP,
            FOREIGN KEY (user_mxid) REFERENCES "user"(mxid) ON DELETE CASCADE ON UPDATE CASCADE,
            FOREIGN KEY (portal_tgid, portal_tg_receiver)
                REFERENCES portal(tgid, tg_receiver) ON DELETE CASCADE
        )
        """
    )



# mautrix-telegram - A Matrix-Telegram puppeting bridge
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Callable

from mautrix.types import EventID, RoomID

from ... import portal as po
from .. import SECTION_PORTAL_MANAGEMENT, CommandEvent, command_handler
from .util import user_has_power_level


async def _get_portal_and_check_permission(evt: CommandEvent) -> po.Portal | None:
    room_id = RoomID(evt.args[0]) if len(evt.args) > 0 else evt.room_id

    portal = await po.Portal.get_by_mxid(room_id)
    if not portal:
        that_this = "This" if room_id == evt.room_id else "That"
        await evt.reply(f"{that_this} is not a portal room.")
        return None

    if portal.peer_type == "user":
        if portal.tg_receiver != evt.sender.tgid:
            await evt.reply("You do not have the permissions to unbridge that portal.")
            return None
        return portal

    if not await user_has_power_level(portal.mxid, evt.az.intent, evt.sender, "unbridge"):
        await evt.reply("You do not have the permissions to unbridge that portal.")
        return None
    return portal


def _get_portal_murder_function(
    action: str, room_id: str, function: Callable, command: str, completed_message: str
) -> dict:
    async def post_confirm(confirm) -> EventID | None:
        confirm.sender.command_status = None
        if len(confirm.args) > 0 and confirm.args[0] == f"confirm-{command}":
            await function()
            if confirm.room_id != room_id:
                return await confirm.reply(completed_message)
        else:
            return await confirm.reply(f"{action} cancelled.")
        return None

    return {
        "next": post_confirm,
        "action": action,
    }


@command_handler(
    needs_auth=False,
    needs_puppeting=False,
    help_section=SECTION_PORTAL_MANAGEMENT,
    help_text=(
        "Remove all users from the current portal room and forget the portal. "
        "Only works for group chats; to delete a private chat portal, simply leave the room."
    ),
)
async def delete_portal(evt: CommandEvent) -> EventID | None:
    portal = await _get_portal_and_check_permission(evt)
    if not portal:
        return None

    evt.sender.command_status = _get_portal_murder_function(
        "Portal deletion",
        portal.mxid,
        portal.cleanup_and_delete,
        "delete",
        "Portal successfully deleted.",
    )
    return await evt.reply(
        "Please confirm deletion of portal "
        f"[{portal.alias or portal.mxid}](https://matrix.to/#/{portal.mxid}) "
        f'to Telegram chat "{portal.title}" '
        "by typing `$cmdprefix+sp confirm-delete`"
        "\n\n"
        "**WARNING:** If the bridge bot has the power level to do so, **this "
        "will kick ALL users** in the room. If you just want to remove the "
        "bridge, use `$cmdprefix+sp unbridge` instead."
    )


@command_handler(
    needs_auth=False,
    needs_puppeting=False,
    help_section=SECTION_PORTAL_MANAGEMENT,
    help_text="Remove puppets from the current portal room and forget the portal.",
)
async def unbridge(evt: CommandEvent) -> EventID | None:
    portal = await _get_portal_and_check_permission(evt)
    if not portal:
        return None

    evt.sender.command_status = _get_portal_murder_function(
        "Room unbridging", portal.mxid, portal.unbridge, "unbridge", "Room successfully unbridged."
    )
    return await evt.reply(
        f'Please confirm unbridging chat "{portal.title}" from room '
        f"[{portal.alias or portal.mxid}](https://matrix.to/#/{portal.mxid}) "
        "by typing `$cmdprefix+sp confirm-unbridge`"
    )



# mautrix-telegram - A Matrix-Telegram puppeting bridge
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from telethon.errors import (
    AboutTooLongError,
    AuthKeyError,
    FirstNameInvalidError,
    HashInvalidError,
    UsernameInvalidError,
    UsernameNotModifiedError,
    UsernameOccupiedError,
)
from telethon.tl.functions.account import (
    GetAuthorizationsRequest,
    ResetAuthorizationRequest,
    UpdateProfileRequest,
    UpdateUsernameRequest,
)
from telethon.tl.types import Authorization

from mautrix.types import EventID

from .. import SECTION_AUTH, CommandEvent, command_handler


@command_handler(
    needs_auth=True,
    help_section=SECTION_AUTH,
    help_args="<_new username_>",
    help_text="Change your Telegram username.",
)
async def username(evt: CommandEvent) -> EventID:
    if len(evt.args) == 0:
        return await evt.reply("**Usage:** `$cmdprefix+sp username <new username>`")
    if evt.sender.is_bot:
        return await evt.reply("Bots can't set their own username.")
    new_name = evt.args[0]
    if new_name == "-":
        new_name = ""
    try:
        await evt.sender.client(UpdateUsernameRequest(username=new_name))
    except UsernameInvalidError:
        return await evt.reply(
            "Invalid username. Usernames must be between 5 and 30 alphanumeric characters."
        )
    except UsernameNotModifiedError:
        return await evt.reply("That is your current username.")
    except UsernameOccupiedError:
        return await evt.reply("That username is already in use.")
    await evt.sender.update_info()
    if not evt.sender.tg_username:
        await evt.reply("Username removed")
    else:
        await evt.reply(f"Username changed to {evt.sender.tg_username}")


@command_handler(
    needs_auth=True,
    help_section=SECTION_AUTH,
    help_args="<_new about_>",
    help_text="Change your Telegram about section.",
)
async def about(evt: CommandEvent) -> EventID:
    if len(evt.args) == 0:
        return await evt.reply("**Usage:** `$cmdprefix+sp about <new about>`")
    if evt.sender.is_bot:
        return await evt.reply("Bots can't set their own about section.")
    new_about = " ".join(evt.args)
    if new_about == "-":
        new_about = ""
    try:
        await evt.sender.client(UpdateProfileRequest(about=new_about))
    except AboutTooLongError:
        return await evt.reply("The provided about section is too long")
    return await evt.reply("About section updated")


@command_handler(
    needs_auth=True,
    help_section=SECTION_AUTH,
    help_args="<_new displayname_>",
    help_text="Change your Telegram displayname.",
)
async def displayname(evt: CommandEvent) -> EventID:
    if len(evt.args) == 0:
        return await evt.reply("**Usage:** `$cmdprefix+sp displayname <new displayname>`")
    if evt.sender.is_bot:
        return await evt.reply("Bots can't set their own displayname.")

    first_name, last_name = (
        (evt.args[0], "") if len(evt.args) == 1 else (" ".join(evt.args[:-1]), evt.args[-1])
    )
    try:
        await evt.sender.client(UpdateProfileRequest(first_name=first_name, last_name=last_name))
    except FirstNameInvalidError:
        return await evt.reply("Invalid first name")
    await evt.sender.update_info()
    return await evt.reply("Displayname updated")


def _format_session(sess: Authorization) -> str:
    return (
        f"**{sess.app_name} {sess.app_version}**  \n"
        f"  **Platform:** {sess.device_model} {sess.platform} {sess.system_version}  \n"
        f"  **Active:** {sess.date_active} (created {sess.date_created})  \n"
        f"  **From:** {sess.ip} - {sess.region}, {sess.country}"
    )


@command_handler(
    needs_auth=True,
    help_section=SECTION_AUTH,
    help_args="<`list`|`terminate`> [_hash_]",
    help_text="View or delete other Telegram sessions.",
)
async def session(evt: CommandEvent) -> EventID:
    if len(evt.args) == 0:
        return await evt.reply("**Usage:** `$cmdprefix+sp session <list|terminate> [hash]`")
    elif evt.sender.is_bot:
        return await evt.reply("Bots can't manage their sessions")
    cmd = evt.args[0].lower()
    if cmd == "list":
        res = await evt.sender.client(GetAuthorizationsRequest())
        session_list = res.authorizations
        current = [s for s in session_list if s.current][0]
        current_text = _format_session(current)
        other_text = "\n".join(
            f"* {_format_session(sess)}  \n  **Hash:** {sess.hash}"
            for sess in session_list
            if not sess.current
        )
        return await evt.reply(
            f"### Current session\n"
            f"{current_text}\n"
            f"\n"
            f"### Other active sessions\n"
            f"{other_text}"
        )
    elif cmd == "terminate" and len(evt.args) > 1:
        try:
            session_hash = int(evt.args[1])
        except ValueError:
            return await evt.reply("Hash must be an integer")
        try:
            ok = await evt.sender.client(ResetAuthorizationRequest(hash=session_hash))
        except HashInvalidError:
            return await evt.reply("Invalid session hash.")
        except AuthKeyError as e:
            if e.message == "FRESH_RESET_AUTHORISATION_FORBIDDEN":
                return await evt.reply(
                    "New sessions can't terminate other sessions. Please wait a while."
                )
            raise
        if ok:
            return await evt.reply("Session terminated successfully.")
        else:
            return await evt.reply("Session not found.")
    else:
        return await evt.reply("**Usage:** `$cmdprefix+sp session <list|terminate> [hash]`")



# mautrix-telegram - A Matrix-Telegram puppeting bridge
# Telegram lottie sticker converter
# Copyright (C) 2019 Randall Eramde Lawrence
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Any, Awaitable, Callable
import asyncio.subprocess
import logging
import os
import os.path
import shutil
import tempfile

from attr import dataclass

from mautrix.util import ffmpeg

log: logging.Logger = logging.getLogger("mau.util.tgs")


@dataclass
class ConvertedSticker:
    mime: str
    data: bytes
    thumbnail_mime: str | None = None
    thumbnail_data: bytes | None = None
    width: int = 0
    height: int = 0


Converter = Callable[[bytes, int, int, Any], Awaitable[ConvertedSticker]]
converters: dict[str, Converter] = {}


def abswhich(program: str | None) -> str | None:
    path = shutil.which(program)
    return os.path.abspath(path) if path else None


lottieconverter = abswhich("lottieconverter")


async def _run_lottieconverter(args: tuple[str, ...], input_data: bytes) -> bytes:
    proc = await asyncio.create_subprocess_exec(
        lottieconverter,
        *args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
        stdin=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await proc.communicate(input_data)
    if proc.returncode == 0:
        return stdout
    else:
        err_text = stderr.decode("utf-8") if stderr else f"unknown ({proc.returncode})"
        raise ffmpeg.ConverterError(f"lottieconverter error: {err_text}")


if lottieconverter:

    async def tgs_to_png(file: bytes, width: int, height: int, **_: Any) -> ConvertedSticker:
        frame = 1
        try:
            converted_png = await _run_lottieconverter(
                args=("-", "-", "png", f"{width}x{height}", str(frame)),
                input_data=file,
            )
            return ConvertedSticker("image/png", converted_png)
        except ffmpeg.ConverterError as e:
            log.error(str(e))
            return ConvertedSticker("application/gzip", file)

    async def tgs_to_gif(
        file: bytes, width: int, height: int, fps: int = 25, **_: Any
    ) -> ConvertedSticker:
        try:
            converted_gif = await _run_lottieconverter(
                args=("-", "-", "gif", f"{width}x{height}", str(fps)),
                input_data=file,
            )
            return ConvertedSticker("image/gif", converted_gif)
        except ffmpeg.ConverterError as e:
            log.error(str(e))
            return ConvertedSticker("application/gzip", file)

    converters["png"] = tgs_to_png
    converters["gif"] = tgs_to_gif

if lottieconverter and ffmpeg.ffmpeg_path:

    async def tgs_to_webm(
        file: bytes, width: int, height: int, fps: int = 30, **_: Any
    ) -> ConvertedSticker:
        with tempfile.TemporaryDirectory(prefix="tgs_") as tmpdir:
            file_template = tmpdir + "/out_"
            try:
                await _run_lottieconverter(
                    args=("-", file_template, "pngs", f"{width}x{height}", str(fps)),
                    input_data=file,
                )
                first_frame_name = min(os.listdir(tmpdir))
                with open(f"{tmpdir}/{first_frame_name}", "rb") as first_frame_file:
                    first_frame_data = first_frame_file.read()
                webm_data = await ffmpeg.convert_path(
                    input_args=("-framerate", str(fps), "-pattern_type", "glob"),
                    input_file=f"{file_template}*.png",
                    output_args=("-c:v", "libvpx-vp9", "-pix_fmt", "yuva420p", "-f", "webm"),
                    output_path_override="-",
                    output_extension=None,
                )
                return ConvertedSticker("video/webm", webm_data, "image/png", first_frame_data)
            except ffmpeg.ConverterError as e:
                log.error(str(e))
        return ConvertedSticker("application/gzip", file)

    async def tgs_to_webp(
        file: bytes, width: int, height: int, fps: int = 30, **_: Any
    ) -> ConvertedSticker:
        with tempfile.TemporaryDirectory(prefix="tgs_") as tmpdir:
            file_template = tmpdir + "/out_"
            try:
                await _run_lottieconverter(
                    args=("-", file_template, "pngs", f"{width}x{height}", str(fps)),
                    input_data=file,
                )
                first_frame_name = min(os.listdir(tmpdir))
                with open(f"{tmpdir}/{first_frame_name}", "rb") as first_frame_file:
                    first_frame_data = first_frame_file.read()
                webp_data = await ffmpeg.convert_path(
                    input_args=("-framerate", str(fps), "-pattern_type", "glob"),
                    input_file=f"{file_template}*.png",
                    output_args=("-c:v", "libwebp_anim", "-pix_fmt", "yuva420p", "-f", "webp"),
                    output_path_override="-",
                    output_extension=None,
                )
                return ConvertedSticker("image/webp", webp_data, "image/png", first_frame_data)
            except ffmpeg.ConverterError as e:
                log.error(str(e))
        return ConvertedSticker("application/gzip", file)

    converters["webm"] = tgs_to_webm
    converters["webp"] = tgs_to_webp


async def convert_tgs_to(
    file: bytes, convert_to: str, width: int, height: int, **kwargs: Any
) -> ConvertedSticker:
    if convert_to in converters:
        converter = converters[convert_to]
        converted = await converter(file, width, height, **kwargs)
        converted.width = width
        converted.height = height
        return converted
    elif convert_to != "disable":
        log.warning(f"Unable to convert animated sticker, type {convert_to} not supported")
    return ConvertedSticker("application/gzip", file)



# mautrix-telegram - A Matrix-Telegram puppeting bridge
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from mautrix.types import EventID

from ... import portal as po
from ...types import TelegramID
from .. import SECTION_CREATING_PORTALS, CommandEvent, command_handler
from .util import get_initial_state, user_has_power_level, warn_missing_power


@command_handler(
    help_section=SECTION_CREATING_PORTALS,
    help_args="[_type_]",
    help_text=(
        "Create a Telegram chat of the given type for the current Matrix room. "
        "The type is either `group`, `supergroup` or `channel` (defaults to `supergroup`)."
    ),
)
async def create(evt: CommandEvent) -> EventID:
    type = evt.args[0] if len(evt.args) > 0 else "supergroup"
    if type not in ("chat", "group", "supergroup", "channel"):
        return await evt.reply(
            "**Usage:** `$cmdprefix+sp create ['group'/'supergroup'/'channel']`"
        )

    if await po.Portal.get_by_mxid(evt.room_id):
        return await evt.reply("This is already a portal room.")

    if not await user_has_power_level(evt.room_id, evt.az.intent, evt.sender, "bridge"):
        return await evt.reply("You do not have the permissions to bridge this room.")

    title, about, levels, encrypted = await get_initial_state(evt.az.intent, evt.room_id)
    if not title:
        return await evt.reply("Please set a title before creating a Telegram chat.")

    supergroup = type == "supergroup"
    type = {
        "supergroup": "channel",
        "channel": "channel",
        "chat": "chat",
        "group": "chat",
    }[type]

    portal = po.Portal(
        tgid=TelegramID(0),
        tg_receiver=TelegramID(0),
        peer_type=type,
        mxid=evt.room_id,
        title=title,
        about=about,
        encrypted=encrypted,
    )

    await warn_missing_power(levels, evt)

    try:
        await portal.create_telegram_chat(evt.sender, supergroup=supergroup)
    except ValueError as e:
        await portal.delete()
        return await evt.reply(e.args[0])



# mautrix-telegram - A Matrix-Telegram puppeting bridge
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from html import escape
import logging
import re

from telethon.errors import RPCError
from telethon.helpers import add_surrogate, del_surrogate
from telethon.tl.custom import Message
from telethon.tl.types import (
    Channel,
    InputPeerChannelFromMessage,
    InputPeerUserFromMessage,
    MessageEntityBlockquote,
    MessageEntityBold,
    MessageEntityBotCommand,
    MessageEntityCashtag,
    MessageEntityCode,
    MessageEntityCustomEmoji,
    MessageEntityEmail,
    MessageEntityHashtag,
    MessageEntityItalic,
    MessageEntityMention,
    MessageEntityMentionName,
    MessageEntityPhone,
    MessageEntityPre,
    MessageEntitySpoiler,
    MessageEntityStrike,
    MessageEntityTextUrl,
    MessageEntityUnderline,
    MessageEntityUrl,
    MessageFwdHeader,
    PeerChannel,
    PeerChat,
    PeerUser,
    SponsoredMessage,
    TypeMessageEntity,
    User,
)

from mautrix.types import Format, MessageType, TextMessageEventContent

from .. import abstract_user as au, portal as po, puppet as pu, user as u
from ..db import Message as DBMessage, TelegramFile as DBTelegramFile
from ..tgclient import MautrixTelegramClient
from ..types import TelegramID
from ..util.file_transfer import UnicodeCustomEmoji, transfer_custom_emojis_to_matrix

log: logging.Logger = logging.getLogger("mau.fmt.tg")


async def _get_fwd_entity(client: MautrixTelegramClient, evt: Message) -> Channel | User | None:
    try:
        return await client.get_entity(evt.fwd_from.from_id)
    except (ValueError, RPCError) as e:
        try:
            input_peer = await client.get_input_entity(evt.peer_id)
            if isinstance(evt.fwd_from.from_id, PeerUser):
                return await client.get_entity(
                    InputPeerUserFromMessage(
                        peer=input_peer, msg_id=evt.id, user_id=evt.fwd_from.from_id.user_id
                    )
                )
            elif isinstance(evt.fwd_from.from_id, PeerChannel):
                return await client.get_entity(
                    InputPeerChannelFromMessage(
                        peer=input_peer, msg_id=evt.id, channel_id=evt.fwd_from.from_id.channel_id
                    )
                )
        except (ValueError, RPCError) as e:
            pass
        return None


async def _add_forward_header(
    client: MautrixTelegramClient, content: TextMessageEventContent, evt: Message
) -> None:
    fwd_from = evt.fwd_from
    fwd_from_html, fwd_from_text = None, None
    if isinstance(fwd_from.from_id, PeerUser):
        user = await u.User.get_by_tgid(TelegramID(fwd_from.from_id.user_id))
        if user:
            fwd_from_text = user.displayname or user.mxid
            fwd_from_html = (
                f"<a href='https://matrix.to/#/{user.mxid}'>{escape(fwd_from_text)}</a>"
            )

        if not fwd_from_text:
            puppet = await pu.Puppet.get_by_peer(fwd_from.from_id, create=False)
            if puppet and puppet.displayname:
                fwd_from_text = puppet.displayname or puppet.mxid
                fwd_from_html = (
                    f"<a href='https://matrix.to/#/{puppet.mxid}'>{escape(fwd_from_text)}</a>"
                )

        if not fwd_from_text:
            user = await _get_fwd_entity(client, evt)
            if user:
                fwd_from_text, _ = pu.Puppet.get_displayname(user, False)
                fwd_from_html = f"<b>{escape(fwd_from_text)}</b>"
            else:
                fwd_from_text = fwd_from_html = "unknown user"
    elif isinstance(fwd_from.from_id, (PeerChannel, PeerChat)):
        from_id = (
            fwd_from.from_id.chat_id
            if isinstance(fwd_from.from_id, PeerChat)
            else fwd_from.from_id.channel_id
        )
        portal = await po.Portal.get_by_tgid(TelegramID(from_id))
        if portal and portal.title:
            fwd_from_text = portal.title
            if portal.alias:
                fwd_from_html = (
                    f"<a href='https://matrix.to/#/{portal.alias}'>{escape(fwd_from_text)}</a>"
                )
            else:
                fwd_from_html = f"channel <b>{escape(fwd_from_text)}</b>"
        else:
            channel = await _get_fwd_entity(client, evt)
            if channel:
                fwd_from_text = f"channel {channel.title}"
                fwd_from_html = f"channel <b>{escape(channel.title)}</b>"
            else:
                fwd_from_text = fwd_from_html = "unknown channel"
    elif fwd_from.from_name:
        fwd_from_text = fwd_from.from_name
        fwd_from_html = f"<b>{escape(fwd_from.from_name)}</b>"
    else:
        fwd_from_text = "unknown source"
        fwd_from_html = f"unknown source"

    content.ensure_has_html()
    content.body = "\n".join([f"> {line}" for line in content.body.split("\n")])
    content.body = f"Forwarded from {fwd_from_text}:\n{content.body}"
    content.formatted_body = (
        f"Forwarded message from {fwd_from_html}<br/>"
        f"<tg-forward><blockquote>{content.formatted_body}</blockquote></tg-forward>"
    )


class ReuploadedCustomEmoji(MessageEntityCustomEmoji):
    file: DBTelegramFile

    def __init__(self, parent: MessageEntityCustomEmoji, file: DBTelegramFile) -> None:
        super().__init__(parent.offset, parent.length, parent.document_id)
        self.file = file


async def _convert_custom_emoji(
    source: au.AbstractUser,
    entities: list[TypeMessageEntity],
    client: MautrixTelegramClient | None = None,
) -> None:
    emoji_ids = [
        entity.document_id for entity in entities if isinstance(entity, MessageEntityCustomEmoji)
    ]
    custom_emojis = await transfer_custom_emojis_to_matrix(source, emoji_ids, client=client)
    if len(custom_emojis) > 0:
        for i, entity in enumerate(entities):
            if isinstance(entity, MessageEntityCustomEmoji):
                entities[i] = ReuploadedCustomEmoji(entity, custom_emojis[entity.document_id])


async def telegram_text_to_matrix_html(
    source: au.AbstractUser,
    text: str,
    entities: list[TypeMessageEntity],
    client: MautrixTelegramClient | None = None,
) -> str:
    if not entities:
        return escape(text).replace("\n", "<br/>")
    await _convert_custom_emoji(source, entities, client=client)
    text = add_surrogate(text)
    html = await _telegram_entities_to_matrix_catch(text, entities)
    html = del_surrogate(html)
    return html


async def telegram_to_matrix(
    evt: Message | SponsoredMessage,
    source: au.AbstractUser,
    client: MautrixTelegramClient | None = None,
    override_text: str = None,
    override_entities: list[TypeMessageEntity] = None,
    require_html: bool = False,
) -> TextMessageEventContent:
    if not client:
        client = source.client
    content = TextMessageEventContent(
        msgtype=MessageType.TEXT,
        body=override_text or evt.message,
    )
    entities = override_entities or evt.entities
    if entities:
        content.format = Format.HTML
        content.formatted_body = await telegram_text_to_matrix_html(
            source, content.body, entities, client=client
        )

    if require_html:
        content.ensure_has_html()

    if getattr(evt, "fwd_from", None):
        await _add_forward_header(client, content, evt)

    if isinstance(evt, Message) and evt.post and evt.post_author:
        content.ensure_has_html()
        content.body += f"\n- {evt.post_author}"
        content.formatted_body += f"<br/><i>- <u>{evt.post_author}</u></i>"

    return content


async def _telegram_entities_to_matrix_catch(text: str, entities: list[TypeMessageEntity]) -> str:
    try:
        return await _telegram_entities_to_matrix(text, entities)
    except Exception:
        log.exception(
            "Failed to convert Telegram format:\nmessage=%s\nentities=%s", text, entities
        )
    return "[failed conversion in _telegram_entities_to_matrix]"


def within_surrogate(text, index):
    """
    `True` if ``index`` is within a surrogate (before and after it, not at!).
    """
    return (
        1 < index < len(text)  # in bounds
        and "\ud800" <= text[index - 1] <= "\udbff"  # current is low surrogate
        and "\udc00" <= text[index] <= "\udfff"  # previous is high surrogate
    )


async def _telegram_entities_to_matrix(
    text: str,
    entities: list[TypeMessageEntity | ReuploadedCustomEmoji],
    offset: int = 0,
    length: int = None,
    in_codeblock: bool = False,
) -> str:
    def text_to_html(
        val: str, _in_codeblock: bool = in_codeblock, escape_html: bool = True
    ) -> str:
        if escape_html:
            val = escape(val)
        if not _in_codeblock:
            val = val.replace("\n", "<br/>")
        return val

    if not entities:
        return text_to_html(text)
    if length is None:
        length = len(text)
    html = []
    last_offset = 0
    for i, entity in enumerate(entities):
        if entity.offset >= offset + length:
            break
        relative_offset = entity.offset - offset
        if relative_offset > last_offset:
            html.append(text_to_html(text[last_offset:relative_offset]))
        elif relative_offset < last_offset:
            continue

        while within_surrogate(text, relative_offset):
            relative_offset += 1
        while within_surrogate(text, relative_offset + entity.length):
            entity.length += 1

        skip_entity = False
        is_code_entity = isinstance(entity, (MessageEntityCode, MessageEntityPre))
        entity_text = await _telegram_entities_to_matrix(
            text=text[relative_offset : relative_offset + entity.length],
            entities=entities[i + 1 :],
            offset=entity.offset,
            length=entity.length,
            in_codeblock=is_code_entity,
        )
        entity_text = text_to_html(entity_text, is_code_entity, escape_html=False)
        entity_type = type(entity)

        if entity_type == MessageEntityBold:
            html.append(f"<strong>{entity_text}</strong>")
        elif entity_type == MessageEntityItalic:
            html.append(f"<em>{entity_text}</em>")
        elif entity_type == MessageEntityUnderline:
            html.append(f"<u>{entity_text}</u>")
        elif entity_type == MessageEntityStrike:
            html.append(f"<del>{entity_text}</del>")
        elif entity_type == MessageEntityBlockquote:
            html.append(f"<blockquote>{entity_text}</blockquote>")
        elif entity_type == MessageEntityCode:
            html.append(
                f"<pre><code>{entity_text}</code></pre>"
                if "\n" in entity_text
                else f"<code>{entity_text}</code>"
            )
        elif entity_type == MessageEntityPre:
            skip_entity = _parse_pre(html, entity_text, entity.language)
        elif entity_type == MessageEntityMention:
            skip_entity = await _parse_mention(html, entity_text)
        elif entity_type == MessageEntityMentionName:
            skip_entity = await _parse_name_mention(html, entity_text, TelegramID(entity.user_id))
        elif entity_type == MessageEntityEmail:
            html.append(f"<a href='mailto:{entity_text}'>{entity_text}</a>")
        elif entity_type in (MessageEntityTextUrl, MessageEntityUrl):
            await _parse_url(
                html, entity_text, entity.url if entity_type == MessageEntityTextUrl else None
            )
        elif entity_type == MessageEntityCustomEmoji:
            html.append(entity_text)
        elif entity_type == ReuploadedCustomEmoji:
            if isinstance(entity.file, UnicodeCustomEmoji):
                html.append(entity.file.emoji)
            else:
                html.append(
                    f"<img data-mx-emoticon data-mau-animated-emoji"
                    f' src="{escape(entity.file.mxc)}" height="32" width="32"'
                    f' alt="{entity_text}" title="{entity_text}"/>'
                )
        elif entity_type in (
            MessageEntityBotCommand,
            MessageEntityHashtag,
            MessageEntityCashtag,
            MessageEntityPhone,
        ):
            html.append(f"<font color='#3771bb'>{entity_text}</font>")
        elif entity_type == MessageEntitySpoiler:
            html.append(f"<span data-mx-spoiler>{entity_text}</span>")
        else:
            skip_entity = True
        last_offset = relative_offset + (0 if skip_entity else entity.length)
    html.append(text_to_html(text[last_offset:]))

    html_string = "".join(html)
    # Remove redundant <br>'s after block tags
    html_string = html_string.replace("</blockquote><br/>", "</blockquote>")
    html_string = html_string.replace("</pre><br/>", "</pre>")
    return html_string


def _parse_pre(html: list[str], entity_text: str, language: str) -> bool:
    if language:
        html.append(f"<pre><code class='language-{language}'>{entity_text}</code></pre>")
    else:
        html.append(f"<pre><code>{entity_text}</code></pre>")
    return False


async def _parse_mention(html: list[str], entity_text: str) -> bool:
    username = entity_text[1:]

    mxid = None
    portal = None
    # This is a bit complicated because public channels have both Puppet and Portal instances.
    # Basically the currently intended output is:
    # User/bot mention (bridge user)          -> real user mention
    # User/bot mention (normal Telegram user) -> ghost user mention
    # Public channel with existing portal     -> room mention
    # Public channel without portal           -> ghost user mention
    # Other chat                              -> room mention
    user = await u.User.find_by_username(username) or await pu.Puppet.find_by_username(username)
    if user:
        if isinstance(user, pu.Puppet) and user.is_channel:
            portal = await po.Portal.get_by_tgid(user.tgid)
        mxid = user.mxid
    else:
        portal = await po.Portal.find_by_username(username)
    if portal and (portal.mxid or not user):
        mxid = portal.alias or portal.mxid

    if mxid:
        html.append(f"<a href='https://matrix.to/#/{mxid}'>{entity_text}</a>")
    else:
        return True
    return False


async def _parse_name_mention(html: list[str], entity_text: str, user_id: TelegramID) -> bool:
    user = await u.User.get_by_tgid(user_id)
    if user:
        mxid = user.mxid
    else:
        puppet = await pu.Puppet.get_by_tgid(user_id, create=False)
        mxid = puppet.mxid if puppet else None
    if mxid:
        html.append(f"<a href='https://matrix.to/#/{mxid}'>{entity_text}</a>")
    else:
        return True
    return False


message_link_regex = re.compile(
    r"https?://t(?:elegram)?\.(?:me|dog)"
    # /username or /c/id
    r"/([A-Za-z][A-Za-z0-9_]{3,31}[A-Za-z0-9]|[Cc]/[0-9]{1,20})"
    # /messageid
    r"/([0-9]{1,20})"
)


async def _parse_url(html: list[str], entity_text: str, url: str) -> None:
    url = escape(url) if url else entity_text
    if not url.startswith(("https://", "http://", "ftp://", "magnet://")):
        url = "http://" + url

    message_link_match = message_link_regex.match(url)
    if message_link_match:
        group, msgid_str = message_link_match.groups()
        msgid = int(msgid_str)

        if group.lower().startswith("c/"):
            portal = await po.Portal.get_by_tgid(TelegramID(int(group[2:])))
        else:
            portal = await po.Portal.find_by_username(group)
        if portal:
            message = await DBMessage.get_one_by_tgid(TelegramID(msgid), portal.tgid)
            if message:
                url = f"https://matrix.to/#/{portal.mxid}/{message.mxid}"

    html.append(f"<a href='{url}'>{entity_text}</a>")



# mautrix-telegram - A Matrix-Telegram puppeting bridge
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from abc import abstractmethod
import abc
import asyncio
import logging

from aiohttp import web
from telethon.errors import (
    AccessTokenExpiredError,
    AccessTokenInvalidError,
    FloodWaitError,
    PasswordEmptyError,
    PasswordHashInvalidError,
    PhoneCodeExpiredError,
    PhoneCodeInvalidError,
    PhoneNumberAppSignupForbiddenError,
    PhoneNumberBannedError,
    PhoneNumberFloodError,
    PhoneNumberInvalidError,
    PhoneNumberUnoccupiedError,
    SessionPasswordNeededError,
    SessionRevokedError,
)

from mautrix.bridge import InvalidAccessToken, OnlyLoginSelf
from mautrix.util import background_task
from mautrix.util.format_duration import format_duration

from ...commands.telegram.auth import enter_password
from ...puppet import Puppet
from ...user import User


class AuthAPI(abc.ABC):
    log: logging.Logger = logging.getLogger("mau.web.auth")
    loop: asyncio.AbstractEventLoop

    def __init__(self, loop: asyncio.AbstractEventLoop):
        self.loop = loop

    @abstractmethod
    def get_login_response(
        self,
        status: int = 200,
        state: str = "",
        username: str = "",
        phone: str = "",
        human_tg_id: str = "",
        mxid: str = "",
        message: str = "",
        error: str = "",
        errcode: str = "",
    ) -> web.Response:
        raise NotImplementedError()

    @abstractmethod
    def get_mx_login_response(
        self,
        status: int = 200,
        state: str = "",
        username: str = "",
        phone: str = "",
        human_tg_id: str = "",
        mxid: str = "",
        message: str = "",
        error: str = "",
        errcode: str = "",
    ) -> web.Response:
        raise NotImplementedError()

    async def post_matrix_token(self, user: User, token: str) -> web.Response:
        puppet = await Puppet.get_by_tgid(user.tgid)
        if puppet.is_real_user:
            return self.get_mx_login_response(
                state="already-logged-in",
                status=409,
                error="You have already logged in with your Matrix account.",
                errcode="already-logged-in",
            )

        try:
            await puppet.switch_mxid(token.strip(), user.mxid)
        except OnlyLoginSelf:
            return self.get_mx_login_response(
                status=403,
                errcode="only-login-self",
                error="You can only log in as your own Matrix user.",
            )
        except InvalidAccessToken:
            return self.get_mx_login_response(
                status=401, errcode="invalid-access-token", error="Failed to verify access token."
            )
        return self.get_mx_login_response(mxid=user.mxid, status=200, state="logged-in")

    async def post_matrix_password(self, user: User, password: str) -> web.Response:
        return self.get_mx_login_response(
            mxid=user.mxid, status=501, error="Not yet implemented", errcode="not-yet-implemented"
        )

    async def post_login_phone(self, user: User, phone: str) -> web.Response:
        if not phone or not phone.strip():
            return self.get_login_response(
                mxid=user.mxid,
                state="request",
                status=400,
                errcode="phone_number_invalid",
                error="Phone number not given.",
            )
        try:
            await user.client.sign_in(phone.strip())
            return self.get_login_response(
                mxid=user.mxid,
                state="code",
                status=200,
                message=(
                    "Code requested successfully. Check your SMS "
                    "or Telegram app and enter the code below."
                ),
            )
        except PhoneNumberInvalidError:
            return self.get_login_response(
                mxid=user.mxid,
                state="request",
                status=400,
                errcode="phone_number_invalid",
                error="Invalid phone number.",
            )
        except PhoneNumberBannedError:
            return self.get_login_response(
                mxid=user.mxid,
                state="request",
                status=403,
                errcode="phone_number_banned",
                error="Your phone number is banned from Telegram.",
            )
        except PhoneNumberAppSignupForbiddenError:
            return self.get_login_response(
                mxid=user.mxid,
                state="request",
                status=403,
                errcode="phone_number_app_signup_forbidden",
                error="You have disabled 3rd party apps on your account.",
            )
        except PhoneNumberUnoccupiedError:
            return self.get_login_response(
                mxid=user.mxid,
                state="request",
                status=404,
                errcode="phone_number_unoccupied",
                error="That phone number has not been registered.",
            )
        except PhoneNumberFloodError:
            return self.get_login_response(
                mxid=user.mxid,
                state="request",
                status=429,
                errcode="phone_number_flood",
                error=(
                    "Your phone number has been temporarily blocked for flooding. "
                    "The ban is usually applied for around a day."
                ),
            )
        except FloodWaitError as e:
            return self.get_login_response(
                mxid=user.mxid,
                state="request",
                status=429,
                errcode="flood_wait",
                error=(
                    "Your phone number has been temporarily blocked for flooding. "
                    f"Please wait for {format_duration(e.seconds)} before trying again."
                ),
            )
        except Exception:
            self.log.exception("Error requesting phone code")
            return self.get_login_response(
                mxid=user.mxid,
                state="request",
                status=500,
                errcode="unknown_error",
                error="Internal server error while requesting code.",
            )

    async def postprocess_login(self, user: User, user_info) -> None:
        existing_user = await User.get_by_tgid(user_info.id)
        if existing_user and existing_user != user:
            await existing_user.log_out()
        background_task.create(user.post_login(user_info, first_login=True))
        if user.command_status and user.command_status["action"] == "Login":
            user.command_status = None

    async def post_login_token(self, user: User, token: str) -> web.Response:
        try:
            user_info = await user.client.sign_in(bot_token=token.strip())
            await self.postprocess_login(user, user_info)
            return self.get_login_response(
                mxid=user.mxid,
                state="logged-in",
                status=200,
                username=user_info.username,
                phone=None,
                human_tg_id=f"@{user_info.username}",
            )
        except AccessTokenInvalidError:
            return self.get_login_response(
                mxid=user.mxid,
                state="token",
                status=401,
                errcode="bot_token_invalid",
                error="Bot token invalid.",
            )
        except AccessTokenExpiredError:
            return self.get_login_response(
                mxid=user.mxid,
                state="token",
                status=403,
                errcode="bot_token_expired",
                error="Bot token expired.",
            )
        except Exception:
            self.log.exception("Error sending bot token")
            return self.get_login_response(
                mxid=user.mxid,
                state="token",
                status=500,
                error="Internal server error while sending token.",
            )

    async def post_login_code(
        self, user: User, code: int, password_in_data: bool
    ) -> web.Response | None:
        if not code:
            return self.get_login_response(
                mxid=user.mxid,
                state="code",
                status=400,
                errcode="phone_code_missing",
                error="You must provide the code from your phone.",
            )
        try:
            user_info = await user.client.sign_in(code=code)
            await self.postprocess_login(user, user_info)
            human_tg_id = f"@{user_info.username}" if user_info.username else f"+{user_info.phone}"
            return self.get_login_response(
                mxid=user.mxid,
                state="logged-in",
                status=200,
                username=user_info.username,
                phone=user_info.phone,
                human_tg_id=human_tg_id,
            )
        except PhoneCodeInvalidError:
            return self.get_login_response(
                mxid=user.mxid,
                state="code",
                status=401,
                errcode="phone_code_invalid",
                error="Incorrect phone code.",
            )
        except PhoneCodeExpiredError:
            return self.get_login_response(
                mxid=user.mxid,
                state="code",
                status=403,
                errcode="phone_code_expired",
                error="Phone code expired.",
            )
        except PhoneNumberUnoccupiedError:
            return self.get_login_response(
                mxid=user.mxid,
                state="code",
                status=403,
                errcode="phone_number_unoccupied",
                error="That phone number has not been registered.",
            )
        except FloodWaitError as e:
            return self.get_login_response(
                mxid=user.mxid,
                state="code",
                status=429,
                errcode="flood_wait",
                error=(
                    "You tried to enter your phone code too many times. "
                    f"Please wait for {format_duration(e.seconds)} before trying again."
                ),
            )
        except SessionPasswordNeededError:
            if not password_in_data:
                if user.command_status and user.command_status["action"] == "Login":
                    user.command_status = {
                        "next": enter_password,
                        "action": "Login (password entry)",
                    }
                message = (
                    "Code accepted, but you have 2-factor "
                    "authentication enabled. Please enter your password."
                )
                return self.get_login_response(
                    mxid=user.mxid, state="password", status=202, message=message
                )
            return None
        except Exception:
            self.log.exception("Error sending phone code")
            return self.get_login_response(
                mxid=user.mxid,
                state="code",
                status=500,
                errcode="unknown_error",
                error="Internal server error while sending code.",
            )

    async def post_login_password(self, user: User, password: str) -> web.Response:
        try:
            user_info = await user.client.sign_in(password=password.strip())
            await self.postprocess_login(user, user_info)
            human_tg_id = f"@{user_info.username}" if user_info.username else f"+{user_info.phone}"
            return self.get_login_response(
                mxid=user.mxid,
                state="logged-in",
                status=200,
                username=user_info.username,
                phone=user_info.phone,
                human_tg_id=human_tg_id,
            )
        except PasswordEmptyError:
            return self.get_login_response(
                mxid=user.mxid,
                state="password",
                status=400,
                errcode="password_empty",
                error="Empty password.",
            )
        except PasswordHashInvalidError:
            return self.get_login_response(
                mxid=user.mxid,
                state="password",
                status=401,
                errcode="password_invalid",
                error="Incorrect password.",
            )
        except SessionRevokedError:
            return self.get_login_response(
                mxid=user.mxid,
                state="request",
                status=401,
                errcode="session_revoked",
                error=(
                    "Please try again. Login cancelled because your other sessions were "
                    "terminated via the Telegram app."
                ),
            )
        except FloodWaitError as e:
            return self.get_login_response(
                mxid=user.mxid,
                state="password",
                status=429,
                errcode="flood_wait",
                error=(
                    "You tried to enter your password too many times. "
                    f"Please wait for {format_duration(e.seconds)} before trying again."
                ),
            )
        except Exception as e:
            self.log.exception("Error sending password")
            if isinstance(e, ValueError) and "You must provide a phone and a code" in str(e):
                return self.get_login_response(
                    mxid=user.mxid,
                    state="request",
                    status=400,
                    errcode="phone_code_not_entered",
                    error="Please request a new phone code and enter it first.",
                )
            return self.get_login_response(
                mxid=user.mxid,
                state="password",
                status=500,
                errcode="unknown_error",
                error=f"Internal server error while sending password. {e}",
            )



# mautrix-telegram - A Matrix-Telegram puppeting bridge
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import base64
import html

from telethon.tl.functions.channels import GetSponsoredMessagesRequest
from telethon.tl.types import Channel, InputChannel, PeerChannel, PeerUser, SponsoredMessage, User
from telethon.tl.types.messages import SponsoredMessages, SponsoredMessagesEmpty

from mautrix.types import MessageType, TextMessageEventContent

from .. import user as u
from ..formatter import telegram_to_matrix


async def get_sponsored_message(
    user: u.User,
    entity: InputChannel,
) -> tuple[SponsoredMessage | None, int | None, Channel | User | None]:
    resp = await user.client(GetSponsoredMessagesRequest(entity))
    if isinstance(resp, SponsoredMessagesEmpty):
        return None, None, None
    assert isinstance(resp, SponsoredMessages)
    msg = resp.messages[0]
    if isinstance(msg.from_id, PeerUser):
        entities = resp.users
        target_id = msg.from_id.user_id
    else:
        entities = resp.chats
        target_id = msg.from_id.channel_id
    try:
        entity = next(ent for ent in entities if ent.id == target_id)
    except StopIteration:
        entity = None
    return msg, target_id, entity


async def make_sponsored_message_content(
    source: u.User, msg: SponsoredMessage, entity: Channel | User
) -> TextMessageEventContent | None:
    content = await telegram_to_matrix(msg, source, require_html=True)
    content.external_url = f"https://t.me/{entity.username}"
    content.msgtype = MessageType.NOTICE
    sponsored_meta = {
        "random_id": base64.b64encode(msg.random_id).decode("utf-8"),
    }
    if isinstance(msg.from_id, PeerChannel):
        sponsored_meta["channel_id"] = msg.from_id.channel_id
        if getattr(msg, "channel_post", None) is not None:
            sponsored_meta["channel_post"] = msg.channel_post
            content.external_url += f"/{msg.channel_post}"
            action = "View Post"
        else:
            action = "View Channel"
    elif isinstance(msg.from_id, PeerUser):
        sponsored_meta["bot_id"] = msg.from_id.user_id
        if msg.start_param:
            content.external_url += f"?start={msg.start_param}"
        action = "View Bot"
    else:
        return None

    if isinstance(entity, User):
        name_parts = [entity.first_name, entity.last_name]
        sponsor_name = " ".join(x for x in name_parts if x)
        sponsor_name_html = f"<strong>{html.escape(sponsor_name)}</strong>"
    elif isinstance(entity, Channel):
        sponsor_name = entity.title
        sponsor_name_html = f"<strong>{html.escape(sponsor_name)}</strong>"
    else:
        sponsor_name = sponsor_name_html = "unknown entity"

    content["fi.mau.telegram.sponsored"] = sponsored_meta
    content.formatted_body += (
        f"<br/><br/>Sponsored message from {sponsor_name_html} "
        f"- <a href='{content.external_url}'>{action}</a>"
    )
    content.body += (
        f"\n\nSponsored message from {sponsor_name} - {action} at {content.external_url}"
    )

    return content



# mautrix-telegram - A Matrix-Telegram puppeting bridge
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

import logging

from mautrix.util import ffmpeg

from .tgs_converter import ConvertedSticker

log: logging.Logger = logging.getLogger("mau.util.webm")


converter_args = {
    "gif": {
        "output_args": ("-vf", "split[s0][s1];[s0]palettegen[p];[s1][p]paletteuse"),
    },
    "png": {
        "input_args": ("-ss", "0"),
        "output_args": ("-frames:v", "1"),
    },
    "webp": {},
}


async def convert_webm_to(file: bytes, convert_to: str) -> ConvertedSticker:
    if convert_to in ("png", "gif", "webp"):
        try:
            converted_data = await ffmpeg.convert_bytes(
                data=file,
                output_extension=f".{convert_to}",
                **converter_args[convert_to],
            )
            return ConvertedSticker(f"image/{convert_to}", converted_data)
        except ffmpeg.ConverterError as e:
            log.error(str(e))
    elif convert_to != "disable":
        log.warning(f"Unable to convert webm animated sticker, type {convert_to} not supported")
    return ConvertedSticker("video/webm", file)



# mautrix-telegram - A Matrix-Telegram puppeting bridge
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from asyncio import Lock
from collections import defaultdict

from mautrix.types import EventID

from ..types import TelegramID


class FakeLock:
    async def __aenter__(self) -> None:
        pass

    async def __aexit__(self, exc_type, exc, tb) -> None:
        pass


class PortalSendLock:
    _send_locks: dict[int, Lock]
    _noop_lock: Lock = FakeLock()

    def __init__(self) -> None:
        self._send_locks = {}

    def __call__(self, user_id: TelegramID, required: bool = True) -> Lock:
        if user_id is None and required:
            raise ValueError("Required send lock for none id")
        try:
            return self._send_locks[user_id]
        except KeyError:
            return self._send_locks.setdefault(user_id, Lock()) if required else self._noop_lock


class PortalReactionLock:
    _reaction_locks: dict[EventID, Lock]

    def __init__(self) -> None:
        self._reaction_locks = defaultdict(lambda: Lock())

    def __call__(self, mxid: EventID) -> Lock:
        return self._reaction_locks[mxid]



# mautrix-telegram - A Matrix-Telegram puppeting bridge
# Copyright (C) 2021 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import Any, Generator, Tuple, Union
from collections import deque
import hashlib

from telethon.tl.patched import Message, MessageService
from telethon.tl.types import (
    Message,
    MessageMediaContact,
    MessageMediaDice,
    MessageMediaDocument,
    MessageMediaGame,
    MessageMediaGeo,
    MessageMediaPhoto,
    MessageMediaPoll,
    MessageMediaUnsupported,
    MessageService,
    PeerChannel,
    PeerChat,
    PeerUser,
    TypeUpdates,
    UpdateNewChannelMessage,
    UpdateNewMessage,
    UpdateShortChatMessage,
    UpdateShortMessage,
)

from mautrix.types import EventID

from .. import portal as po
from ..types import TelegramID

DedupMXID = Tuple[EventID, TelegramID]
TypeMessage = Union[Message, MessageService, UpdateShortMessage, UpdateShortChatMessage]

media_content_table = {
    MessageMediaContact: lambda media: [media.user_id],
    MessageMediaDocument: lambda media: [media.document.id],
    MessageMediaPhoto: lambda media: [media.photo.id if media.photo else 0],
    MessageMediaGeo: lambda media: [media.geo.long, media.geo.lat],
    MessageMediaGame: lambda media: [media.game.id],
    MessageMediaPoll: lambda media: [media.poll.id],
    MessageMediaDice: lambda media: [media.value, media.emoticon],
    MessageMediaUnsupported: lambda media: ["unsupported media"],
}


class PortalDedup:
    cache_queue_length: int = 256

    _dedup: deque[bytes | int]
    _dedup_mxid: dict[bytes | int, DedupMXID]
    _dedup_action: deque[bytes | int]
    _portal: po.Portal

    def __init__(self, portal: po.Portal) -> None:
        self._dedup = deque()
        self._dedup_mxid = {}
        self._dedup_action = deque(maxlen=self.cache_queue_length)
        self._portal = portal

    @property
    def _always_force_hash(self) -> bool:
        return self._portal.peer_type == "chat"

    def _hash_content(self, event: TypeMessage) -> Generator[Any, None, None]:
        if not self._always_force_hash:
            yield event.id
        yield int(event.date.timestamp())
        if isinstance(event, MessageService):
            yield event.from_id
            yield event.action
        else:
            yield event.message.strip()
            if event.fwd_from:
                yield event.fwd_from.from_id
            if isinstance(event, Message) and event.media:
                media_hash_func = media_content_table.get(type(event.media)) or (
                    lambda media: ["unknown media"]
                )
                yield media_hash_func(event.media)

    def hash_event(self, event: TypeMessage) -> bytes:
        return hashlib.sha256(
            "-".join(str(a) for a in self._hash_content(event)).encode("utf-8")
        ).digest()

    def check_action(self, event: TypeMessage) -> bool:
        dedup_id = self.hash_event(event) if self._always_force_hash else event.id
        if dedup_id in self._dedup_action:
            return True

        self._dedup_action.appendleft(dedup_id)
        return False

    def update(
        self,
        event: TypeMessage,
        mxid: DedupMXID = None,
        expected_mxid: DedupMXID | None = None,
        force_hash: bool = False,
    ) -> tuple[bytes, DedupMXID | None]:
        evt_hash = self.hash_event(event)
        dedup_id = evt_hash if self._always_force_hash or force_hash else event.id
        try:
            found_mxid = self._dedup_mxid[dedup_id]
        except KeyError:
            return evt_hash, None

        if found_mxid != expected_mxid:
            return evt_hash, found_mxid
        self._dedup_mxid[dedup_id] = mxid
        if evt_hash != dedup_id:
            self._dedup_mxid[evt_hash] = mxid
        return evt_hash, None

    def check(
        self, event: TypeMessage, mxid: DedupMXID = None, force_hash: bool = False
    ) -> tuple[bytes, DedupMXID | None]:
        evt_hash = self.hash_event(event)
        dedup_id = evt_hash if self._always_force_hash or force_hash else event.id
        if dedup_id in self._dedup:
            return evt_hash, self._dedup_mxid[dedup_id]

        self._dedup_mxid[dedup_id] = mxid
        self._dedup.appendleft(dedup_id)
        if evt_hash != dedup_id:
            self._dedup_mxid[evt_hash] = mxid
            self._dedup.appendleft(evt_hash)

        while len(self._dedup) > self.cache_queue_length:
            del self._dedup_mxid[self._dedup.pop()]
        return evt_hash, None

    def register_outgoing_actions(self, response: TypeUpdates) -> None:
        for update in response.updates:
            check_dedup = isinstance(
                update, (UpdateNewMessage, UpdateNewChannelMessage)
            ) and isinstance(update.message, MessageService)
            if check_dedup:
                self.check(update.message)



# mautrix-telegram - A Matrix-Telegram puppeting bridge
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from typing import Any, Literal, TypedDict
from pathlib import Path
import argparse
import asyncio
import io
import json
import logging
import math
import mimetypes
import pickle
import random
import string

from lottie.exporters import export_tgs
from lottie.exporters.cairo import export_png
from lottie.exporters.tgs_validator import Severity, TgsValidator
from lottie.importers.svg import import_svg
from lottie.objects import Animation
from lottie.utils.stripper import float_strip
from PIL import Image
from telethon import TelegramClient
from telethon.custom import Conversation, Message
from telethon.tl.functions.messages import GetStickerSetRequest
from telethon.tl.types import (
    Document,
    DocumentAttributeCustomEmoji,
    DocumentAttributeFilename,
    DocumentAttributeImageSize,
    InputMediaUploadedDocument,
    InputStickerSetShortName,
)
import aiohttp

mimetypes.add_type("image/webp", ".webp")

parser = argparse.ArgumentParser(description="mautrix-telegram unicode emoji packer")
parser.add_argument(
    "-i", "--api-id", type=int, required=True, metavar="<api id>", help="Telegram API ID"
)
parser.add_argument(
    "-a", "--api-hash", type=str, required=True, metavar="<api hash>", help="Telegram API hash"
)
parser.add_argument(
    "-s",
    "--session",
    type=str,
    default="unicodemojipacker.session",
    metavar="<file name>",
    help="Telethon session name",
)
parser.add_argument(
    "-o",
    "--output",
    type=str,
    default="mautrix_telegram/unicodemojipack.json",
    metavar="<file name>",
    help="Path to save created emoji pack document IDs",
)
parser.add_argument(
    "-f",
    "--font-directory",
    type=Path,
    required=True,
    metavar="<directory path>",
    help="Path to the Noto color emoji files",
)
parser.add_argument(
    "-m",
    "--media-directory",
    type=Path,
    required=True,
    metavar="<directory path>",
    help="Path to save converted tgs and webp emoji files",
)
args = parser.parse_args()
font_dir: Path = args.font_directory
media_dir: Path = args.media_directory

EMOJI_DATA_URL = "https://raw.githubusercontent.com/iamcal/emoji-data/master/emoji.json"


def unified_to_unicode(unified: str) -> str:
    return (
        "".join(rf"\U{chunk:0>8}" for chunk in unified.split("-"))
        .encode("ascii")
        .decode("unicode_escape")
    )


def tag_to_str(unified: str) -> str:
    return "".join(chr(int(x.removeprefix("E00"), 16)) for x in unified.split("-"))


EmojiType = Literal["webp", "tgs"]
PackType = Literal["Animated emoji", "Static emoji"]


class Emoji(TypedDict):
    hex: str
    emoji: str
    type: EmojiType
    filename: str


class EmojiData(TypedDict):
    tgs: list[Emoji]
    webp: list[Emoji]


def parse_emoji_data(tone: dict[str, Any], emoji: dict[str, Any]) -> Emoji:
    hex = (tone["non_qualified"] or tone["unified"]).replace("-FE0F", "")
    filename_hex = hex.replace("-", "_").lower()
    filename = f"svg/emoji_u{filename_hex}.svg"
    if emoji["category"] == "Flags" and emoji["subcategory"] in (
        "country-flag",
        "subdivision-flag",
    ):
        filename = f"third_party/region-flags/waved-svg/emoji_u{filename_hex}.svg"

    with (font_dir / filename).open() as f:
        lot: Animation = import_svg(f)
    float_strip(lot)
    lot.tgs_sanitize()

    output = io.BytesIO()
    export_tgs(lot, output)

    validator = TgsValidator()
    validator(lot)
    validator.check_size(len(output.getvalue()))
    errors = [err for err in validator.errors if err.severity != Severity.Note]
    if errors or ("region-flags" in filename and len(output.getvalue()) > 32768):
        lot.scale(100, 100)

        png_out = io.BytesIO()
        export_png(lot, png_out)
        img = Image.open(png_out)
        output = io.BytesIO()
        output.name = "image.webp"
        img.save(output, "webp")

        media_type: EmojiType = "webp"
    else:
        media_type: EmojiType = "tgs"
    path = media_dir / f"{filename_hex}.{media_type}"
    with path.open("wb") as f:
        f.write(output.getvalue())
    print(
        "Converted", filename, "->", path.name, "//" if errors else "", "\n".join(map(str, errors))
    )

    return {
        "hex": hex,
        "emoji": unified_to_unicode(tone["unified"]),
        "type": media_type,
        "filename": path.name,
    }


async def load_emoji_data() -> EmojiData:
    cache_path = media_dir / "conversion-cache.json"
    try:
        with cache_path.open() as f:
            return json.load(f)
    except FileNotFoundError:
        pass
    async with aiohttp.ClientSession() as sess, sess.get(EMOJI_DATA_URL) as resp:
        raw_emoji_data = sorted(
            await resp.json(content_type=None),
            key=lambda dat: dat["sort_order"],
        )
    tgs_emoji = []
    webp_emoji = []
    for emoji in raw_emoji_data:
        for tone in (emoji, *emoji.get("skin_variations", {}).values()):
            parsed_emoji = parse_emoji_data(tone, emoji)
            if parsed_emoji["type"] == "tgs":
                tgs_emoji.append(parsed_emoji)
            else:
                webp_emoji.append(parsed_emoji)
    full_data = {"tgs": tgs_emoji, "webp": webp_emoji}
    with cache_path.open("w") as f:
        json.dump(full_data, f, ensure_ascii=False)
    return full_data


async def create_pack(conv: Conversation, name: str, pack_type: str) -> None:
    await conv.send_message("/newemojipack")
    resp: Message = await conv.get_response()
    assert "A new set of custom emoji" in resp.raw_text
    assert "Please choose the type" in resp.raw_text
    await conv.send_message(pack_type)
    resp = await conv.get_response()
    if pack_type == "Animated emoji":
        assert "When ready to upload, tell me the name of your set." in resp.raw_text
    else:
        assert "Now choose a name for your set." in resp.raw_text
    await conv.send_message(name)
    resp = await conv.get_response()
    if pack_type == "Animated emoji":
        assert "Now send me the first animated emoji" in resp.raw_text
    else:
        assert "Now send me the custom emoji" in resp.raw_text


async def publish_pack(conv: Conversation, shortname: str) -> None:
    await conv.send_message("/publish")

    resp: Message = await conv.get_response()
    assert "You can send me a custom emoji from your emoji set" in resp.raw_text
    await conv.send_message("/skip")

    resp = await conv.get_response()
    assert "Please provide a short name for your emoji set" in resp.raw_text
    await conv.send_message(shortname)

    resp = await conv.get_response()
    assert "I've just published your emoji set" in resp.raw_text


async def send_emoji(
    conv: Conversation, file: bytes | Path | InputMediaUploadedDocument, emoji: str
) -> None:
    await conv.send_file(file)
    resp: Message = await conv.get_response()
    assert "Send me a replacement emoji that corresponds to your custom emoji" in resp.raw_text
    await conv.send_message(emoji)
    resp = await conv.get_response()
    if "Sorry, too many attempts" in resp.raw_text:
        print(resp.raw_text)
        input("Press enter to continue")
        await conv.send_message(emoji)
        resp = await conv.get_response()
    while "Please send an emoji that best describes your custom emoji." in resp.raw_text:
        emoji = input(f"{emoji} was rejected, provide replacement: ")
        await conv.send_message(emoji)
        resp = await conv.get_response()
    assert "Congratulations" in resp.raw_text


class CachedPack(TypedDict):
    name: str
    short_name: str
    part: int
    type: PackType
    published: bool
    collected: bool
    emojis: list[Emoji]


class CachedData(TypedDict):
    packs: list[CachedPack]


def _split_packs_int(
    emoji_list: list[Emoji], pack_type: PackType, current_part: int, total_parts: int
) -> tuple[list[CachedPack], int]:
    packs = []
    current_pack: CachedPack | None = None
    for i, emoji in enumerate(emoji_list):
        if i % 200 == 0:
            current_part += 1
            random_id = "".join(random.choices(string.ascii_lowercase + string.digits, k=8))
            short_name = f"mxtg_unicodemoji_{random_id}"
            name = f"mautrix-telegram unicodemoji ({current_part}/{total_parts})"
            current_pack = {
                "type": pack_type,
                "short_name": short_name,
                "part": current_part,
                "name": name,
                "published": False,
                "collected": False,
                "emojis": [],
            }
            packs.append(current_pack)
        current_pack["emojis"].append(emoji)
    return packs, current_part


def split_packs(emoji_data: EmojiData) -> list[CachedPack]:
    total_parts = math.ceil(len(emoji_data["tgs"]) / 200) + math.ceil(
        len(emoji_data["webp"]) / 200
    )
    current_part = 0
    animated_packs, current_part = _split_packs_int(
        emoji_data["tgs"], "Animated emoji", current_part, total_parts
    )
    static_packs, current_part = _split_packs_int(
        emoji_data["webp"], "Static emoji", current_part, total_parts
    )
    return animated_packs + static_packs


async def create_and_fill_pack(
    client: TelegramClient, conv: Conversation, pack: CachedPack
) -> None:
    if pack["short_name"] == "mxtg_unicodemoji_xvzs6743":
        print("Continuing pack", pack["name"])
    else:
        print("Creating pack", pack["name"])
        await create_pack(conv, pack["name"], pack["type"])
    total = len(pack["emojis"])
    for i, emoji in enumerate(pack["emojis"]):
        if pack["short_name"] == "mxtg_unicodemoji_xvzs6743" and i < 87:
            continue
        print(f"Adding emoji {i+1}/{total}", emoji["hex"], emoji["emoji"])
        emoji_file = media_dir / emoji["filename"]
        if emoji["type"] == "webp":
            attrs = [
                DocumentAttributeImageSize(w=100, h=100),
                DocumentAttributeFilename(file_name="image.webp"),
            ]
            with emoji_file.open("rb") as f:
                file_handle = await client.upload_file(f, file_name="emoji.webp")
            emoji_file = InputMediaUploadedDocument(
                file_handle, mime_type="image/webp", attributes=attrs
            )
        await send_emoji(conv, emoji_file, emoji["emoji"])
        await asyncio.sleep(2)
    print("Publishing pack", pack["short_name"])
    await publish_pack(conv, pack["short_name"])


async def main():
    logging.basicConfig(level=logging.INFO)

    emoji_data = await load_emoji_data()

    split_cache = media_dir / "split-cache.json"
    try:
        with split_cache.open() as f:
            packs: list[CachedPack] = json.load(f)
    except FileNotFoundError:
        packs = split_packs(emoji_data)
        with split_cache.open("w") as f:
            json.dump(packs, f)

    doc_id_file = Path(args.output)
    try:
        with doc_id_file.open() as f:
            doc_ids = json.load(f)
    except FileNotFoundError:
        doc_ids = {}

    client = TelegramClient(args.session, args.api_id, args.api_hash, flood_sleep_threshold=3600)
    await client.start()
    async with client.conversation("Stickers", max_messages=20000) as conv:
        for pack in packs:
            if not pack["published"]:
                await create_and_fill_pack(client, conv, pack)
                pack["published"] = True
                with split_cache.open("w") as f:
                    json.dump(packs, f, ensure_ascii=False)
            if not pack["collected"] or True:
                print("Collecting document IDs from pack", pack["short_name"])
                stickers = await client(
                    GetStickerSetRequest(InputStickerSetShortName(pack["short_name"]), 0)
                )
                doc: Document
                for i, doc in enumerate(stickers.documents):
                    attr = next(
                        attr
                        for attr in doc.attributes
                        if isinstance(attr, DocumentAttributeCustomEmoji)
                    )
                    base_emoji = attr.alt.replace("\ufe0f", "")
                    emoji = pack["emojis"][i]["emoji"].replace("\ufe0f", "")
                    doc_ids[emoji] = doc.id
                    print(f"Mapped {emoji} (fallback: {base_emoji}) -> {doc_ids[emoji]}")
                pack["collected"] = True
                with split_cache.open("w") as f:
                    json.dump(packs, f, ensure_ascii=False)
                with doc_id_file.open("w") as f:
                    json.dump(doc_ids, f, ensure_ascii=False)
                print("Pack completed")
                await asyncio.sleep(5)
    with open(args.output.replace(".json", ".pickle"), "wb") as f:
        pickle.dump(doc_ids, f)
    print("Wrote pickle")


asyncio.run(main())



# mautrix-telegram - A Matrix-Telegram puppeting bridge
# Copyright (C) 2022 Tulir Asokan
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union
from abc import ABC, abstractmethod
import asyncio
import logging
import platform
import time

from telethon.errors import AuthKeyError, UnauthorizedError
from telethon.network import (
    Connection,
    ConnectionTcpFull,
    ConnectionTcpMTProxyRandomizedIntermediate,
)
from telethon.sessions import Session
from telethon.tl.patched import Message, MessageService
from telethon.tl.types import (
    Channel,
    Chat,
    MessageActionChannelMigrateFrom,
    MessageEmpty,
    PeerChannel,
    PeerChat,
    PeerUser,
    PhoneCallRequested,
    TypeUpdate,
    UpdateChannel,
    UpdateChannelUserTyping,
    UpdateChatDefaultBannedRights,
    UpdateChatParticipantAdmin,
    UpdateChatParticipants,
    UpdateChatUserTyping,
    UpdateDeleteChannelMessages,
    UpdateDeleteMessages,
    UpdateEditChannelMessage,
    UpdateEditMessage,
    UpdateFolderPeers,
    UpdateMessageReactions,
    UpdateNewChannelMessage,
    UpdateNewMessage,
    UpdateNotifySettings,
    UpdatePhoneCall,
    UpdatePinnedChannelMessages,
    UpdatePinnedDialogs,
    UpdatePinnedMessages,
    UpdateReadChannelInbox,
    UpdateReadHistoryInbox,
    UpdateReadHistoryOutbox,
    UpdateShort,
    UpdateShortChatMessage,
    UpdateShortMessage,
    UpdateUser,
    UpdateUserName,
    UpdateUserStatus,
    UpdateUserTyping,
    User,
    UserStatusOffline,
    UserStatusOnline,
)

from mautrix.appservice import AppService
from mautrix.errors import MatrixError
from mautrix.types import PresenceState, UserID
from mautrix.util import background_task
from mautrix.util.logging import TraceLogger
from mautrix.util.opt_prometheus import Counter, Histogram

from . import __version__, portal as po, puppet as pu
from .config import Config
from .db import Message as DBMessage, PgSession
from .tgclient import MautrixTelegramClient
from .types import TelegramID

if TYPE_CHECKING:
    from .__main__ import TelegramBridge
    from .bot import Bot

UpdateMessage = Union[
    UpdateShortChatMessage,
    UpdateShortMessage,
    UpdateNewChannelMessage,
    UpdateNewMessage,
    UpdateEditMessage,
    UpdateEditChannelMessage,
]
UpdateMessageContent = Union[
    UpdateShortMessage, UpdateShortChatMessage, Message, MessageService, MessageEmpty
]

UPDATE_TIME = Histogram(
    name="bridge_telegram_update",
    documentation="Time spent processing Telegram updates",
    labelnames=("update_type",),
)
UPDATE_ERRORS = Counter(
    name="bridge_telegram_update_error",
    documentation="Number of fatal errors while handling Telegram updates",
    labelnames=("update_type",),
)


class AbstractUser(ABC):
    loop: asyncio.AbstractEventLoop = None
    log: TraceLogger
    az: AppService
    bridge: "TelegramBridge"
    config: Config
    relaybot: "Bot"
    ignore_incoming_bot_events: bool = True
    max_deletions: int = 10

    client: MautrixTelegramClient | None
    mxid: UserID | None

    tgid: TelegramID | None
    username: str | None
    is_bot: bool

    is_relaybot: bool

    puppet_whitelisted: bool
    whitelisted: bool
    relaybot_whitelisted: bool
    matrix_puppet_whitelisted: bool
    is_admin: bool

    def __init__(self) -> None:
        self.is_admin = False
        self.matrix_puppet_whitelisted = False
        self.puppet_whitelisted = False
        self.whitelisted = False
        self.relaybot_whitelisted = False
        self.client = None
        self.is_relaybot = False
        self.is_bot = False

    @property
    def connected(self) -> bool:
        return self.client and self.client.is_connected()

    @property
    def _proxy_settings(self) -> tuple[type[Connection], tuple[Any, ...] | None]:
        proxy_type = self.config["telegram.proxy.type"].lower()
        connection = ConnectionTcpFull
        connection_data = (
            self.config["telegram.proxy.address"],
            self.config["telegram.proxy.port"],
            self.config["telegram.proxy.rdns"],
            self.config["telegram.proxy.username"],
            self.config["telegram.proxy.password"],
        )
        if proxy_type == "disabled":
            connection_data = None
        elif proxy_type == "socks4":
            connection_data = (1,) + connection_data
        elif proxy_type == "socks5":
            connection_data = (2,) + connection_data
        elif proxy_type == "http":
            connection_data = (3,) + connection_data
        elif proxy_type == "mtproxy":
            connection = ConnectionTcpMTProxyRandomizedIntermediate
            connection_data = (connection_data[0], connection_data[1], connection_data[4])

        return connection, connection_data

    @classmethod
    def init_cls(cls, bridge: "TelegramBridge") -> None:
        cls.bridge = bridge
        cls.config = bridge.config
        cls.loop = bridge.loop
        cls.az = bridge.az
        cls.ignore_incoming_bot_events = cls.config["bridge.relaybot.ignore_own_incoming_events"]
        cls.max_deletions = cls.config["bridge.max_telegram_delete"]

    async def _init_client(self) -> None:
        self.log.debug(f"Initializing client for {self.name}")

        session = await PgSession.get(self.name)
        if self.config["telegram.server.enabled"]:
            session.set_dc(
                self.config["telegram.server.dc"],
                self.config["telegram.server.ip"],
                self.config["telegram.server.port"],
            )

        if self.is_relaybot:
            base_logger = logging.getLogger("telethon.relaybot")
        else:
            base_logger = logging.getLogger(f"telethon.{self.tgid or -hash(self.mxid)}")

        device = self.config["telegram.device_info.device_model"]
        sysversion = self.config["telegram.device_info.system_version"]
        appversion = self.config["telegram.device_info.app_version"]
        connection, proxy = self._proxy_settings
        if proxy:
            self.log.debug(f"Using proxy setting: {proxy}")

        assert isinstance(session, Session)

        self.client = MautrixTelegramClient(
            session=session,
            api_id=self.config["telegram.api_id"],
            api_hash=self.config["telegram.api_hash"],
            app_version=__version__ if appversion == "auto" else appversion,
            system_version=(
                MautrixTelegramClient.__version__ if sysversion == "auto" else sysversion
            ),
            device_model=(
                f"{platform.system()} {platform.release()}" if device == "auto" else device
            ),
            timeout=self.config["telegram.connection.timeout"],
            connection_retries=self.config["telegram.connection.retries"],
            retry_delay=self.config["telegram.connection.retry_delay"],
            flood_sleep_threshold=self.config["telegram.connection.flood_sleep_threshold"],
            request_retries=self.config["telegram.connection.request_retries"],
            connection=connection,
            proxy=proxy,
            raise_last_call_error=True,
            catch_up=self.config["telegram.catch_up"],
            sequential_updates=self.config["telegram.sequential_updates"],
            loop=self.loop,
            base_logger=base_logger,
            update_error_callback=self._telethon_update_error_callback,
            use_ipv6=self.config["telegram.connection.use_ipv6"],
        )
        self.client.add_event_handler(self._update_catch)

    @abstractmethod
    async def on_signed_out(self, err: UnauthorizedError | AuthKeyError) -> None:
        pass

    async def _telethon_update_error_callback(self, err: Exception) -> None:
        if isinstance(err, (UnauthorizedError, AuthKeyError)):
            background_task.create(self.on_signed_out(err))
            return
        if self.config["telegram.exit_on_update_error"]:
            self.log.critical(f"Stopping due to update handling error {type(err).__name__}")
            self.bridge.manual_stop(50)
        else:
            self.log.info("Recreating Telethon connection in 60 seconds")
            await asyncio.sleep(60)
            self.log.debug("Now recreating Telethon connection")
            await self.stop()
            await self.start()

    @abstractmethod
    async def update(self, update: TypeUpdate) -> bool:
        return False

    @abstractmethod
    async def post_login(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def register_portal(self, portal: po.Portal) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def unregister_portal(self, tgid: int, tg_receiver: int) -> None:
        raise NotImplementedError()

    async def _update_catch(self, update: TypeUpdate) -> None:
        start_time = time.time()
        update_type = type(update).__name__
        try:
            if not await self.update(update):
                await self._update(update)
        except Exception:
            self.log.exception("Failed to handle Telegram update")
            UPDATE_ERRORS.labels(update_type=update_type).inc()
        UPDATE_TIME.labels(update_type=update_type).observe(time.time() - start_time)

    @property
    @abstractmethod
    def name(self) -> str:
        raise NotImplementedError()

    async def is_logged_in(self) -> bool:
        return (
            self.client and self.client.is_connected() and await self.client.is_user_authorized()
        )

    async def has_full_access(self, allow_bot: bool = False) -> bool:
        return (
            self.puppet_whitelisted
            and (not self.is_bot or allow_bot)
            and await self.is_logged_in()
        )

    async def start(self, delete_unless_authenticated: bool = False) -> AbstractUser:
        if not self.client:
            await self._init_client()
        attempts = 1
        while True:
            try:
                await self.client.connect()
            except Exception:
                attempts += 1
                if attempts > 10:
                    raise
                self.log.exception("Exception connecting to Telegram, retrying in 5s...")
                await asyncio.sleep(5)
            else:
                break
        self.log.debug(f"{'Bot' if self.is_relaybot else self.mxid} connected: {self.connected}")
        return self

    async def ensure_started(self, even_if_no_session=False) -> AbstractUser:
        if self.connected:
            return self
        session_exists = await PgSession.has(self.mxid)
        if even_if_no_session or session_exists:
            self.log.debug(
                f"Starting client due to ensure_started({even_if_no_session=}, {session_exists=})"
            )
            await self.start(delete_unless_authenticated=not even_if_no_session)
        return self

    async def stop(self) -> None:
        if self.client:
            await self.client.disconnect()
            self.client = None

    # region Telegram update handling

    async def _update(self, update: TypeUpdate) -> None:
        if isinstance(update, UpdateShort):
            update = update.update
        background_task.create(self._handle_entity_updates(getattr(update, "_entities", {})))
        if isinstance(
            update,
            (
                UpdateShortChatMessage,
                UpdateShortMessage,
                UpdateNewChannelMessage,
                UpdateNewMessage,
                UpdateEditMessage,
                UpdateEditChannelMessage,
            ),
        ):
            await self.update_message(update)
        elif isinstance(update, UpdateDeleteMessages):
            await self.delete_message(update)
        elif isinstance(update, UpdateDeleteChannelMessages):
            await self.delete_channel_message(update)
        elif isinstance(update, UpdatePhoneCall):
            await self.update_phone_call(update)
        elif isinstance(update, UpdateMessageReactions):
            await self.update_reactions(update)
        elif isinstance(update, (UpdateChatUserTyping, UpdateChannelUserTyping, UpdateUserTyping)):
            await self.update_typing(update)
        elif isinstance(update, UpdateUserStatus):
            await self.update_status(update)
        elif isinstance(update, UpdateChatParticipantAdmin):
            await self.update_admin(update)
        elif isinstance(update, UpdateChatParticipants):
            await self.update_participants(update)
        elif isinstance(update, UpdateChatDefaultBannedRights):
            await self.update_default_banned_rights(update)
        elif isinstance(update, (UpdatePinnedMessages, UpdatePinnedChannelMessages)):
            await self.update_pinned_messages(update)
        elif isinstance(update, (UpdateUserName, UpdateUser)):
            await self.update_others_info(update)
        elif isinstance(update, UpdateReadHistoryOutbox):
            await self.update_read_receipt(update)
        elif isinstance(update, (UpdateReadHistoryInbox, UpdateReadChannelInbox)):
            await self.update_own_read_receipt(update)
        elif isinstance(update, UpdateFolderPeers):
            await self.update_folder_peers(update)
        elif isinstance(update, UpdatePinnedDialogs):
            await self.update_pinned_dialogs(update)
        elif isinstance(update, UpdateNotifySettings):
            await self.update_notify_settings(update)
        elif isinstance(update, UpdateChannel):
            await self.update_channel(update)
        else:
            self.log.trace("Unhandled update: %s", update)

    async def update_folder_peers(self, update: UpdateFolderPeers) -> None:
        pass

    async def update_pinned_dialogs(self, update: UpdatePinnedDialogs) -> None:
        pass

    async def update_notify_settings(self, update: UpdateNotifySettings) -> None:
        pass

    async def update_pinned_messages(
        self, update: UpdatePinnedMessages | UpdatePinnedChannelMessages
    ) -> None:
        if isinstance(update, UpdatePinnedMessages):
            portal = await po.Portal.get_by_entity(update.peer, tg_receiver=self.tgid)
        else:
            portal = await po.Portal.get_by_tgid(TelegramID(update.channel_id))
        if portal and portal.mxid:
            await portal.receive_telegram_pin_ids(
                update.messages, self.tgid, remove=not update.pinned
            )

    @staticmethod
    async def update_participants(update: UpdateChatParticipants) -> None:
        portal = await po.Portal.get_by_tgid(TelegramID(update.participants.chat_id))
        if portal and portal.mxid:
            await portal.update_power_levels(update.participants.participants)

    @staticmethod
    async def update_default_banned_rights(update: UpdateChatDefaultBannedRights) -> None:
        portal = await po.Portal.get_by_entity(update.peer)
        if portal and portal.mxid:
            await portal.update_default_banned_rights(update.default_banned_rights)

    async def update_read_receipt(self, update: UpdateReadHistoryOutbox) -> None:
        if not isinstance(update.peer, PeerUser):
            self.log.debug("Unexpected read receipt peer: %s", update.peer)
            return

        portal = await po.Portal.get_by_tgid(
            TelegramID(update.peer.user_id), tg_receiver=self.tgid
        )
        if not portal or not portal.mxid:
            return

        # We check that these are user read receipts, so tg_space is always the user ID.
        message = await DBMessage.get_one_by_tgid(
            TelegramID(update.max_id), self.tgid, edit_index=-1
        )
        if not message:
            return

        puppet = await pu.Puppet.get_by_peer(update.peer)
        await puppet.intent.mark_read(portal.mxid, message.mxid)

    async def update_own_read_receipt(
        self, update: UpdateReadHistoryInbox | UpdateReadChannelInbox
    ) -> None:
        puppet = await pu.Puppet.get_by_tgid(self.tgid)
        if not puppet.is_real_user:
            return

        self.log.debug("Handling own read receipt: %s", update)
        if isinstance(update, UpdateReadChannelInbox):
            portal = await po.Portal.get_by_tgid(TelegramID(update.channel_id))
        elif isinstance(update.peer, PeerChat):
            portal = await po.Portal.get_by_tgid(TelegramID(update.peer.chat_id))
        elif isinstance(update.peer, PeerUser):
            portal = await po.Portal.get_by_tgid(
                TelegramID(update.peer.user_id), tg_receiver=self.tgid
            )
        else:
            self.log.debug("Unexpected own read receipt peer: %s", update.peer)
            return

        if not portal or not portal.mxid:
            # TODO This explodes on channels because the field is channel_id
            self.log.debug(f"Dropping own read receipt in unknown chat ({update.peer})")
            return

        tg_space = portal.tgid if portal.peer_type == "channel" else self.tgid
        message = await DBMessage.get_one_by_tgid(
            TelegramID(update.max_id), tg_space, edit_index=-1
        )
        if not message:
            self.log.debug(
                f"Dropping own read receipt: unknown message {update.max_id}@{tg_space}"
            )
            return

        await puppet.intent.mark_read(portal.mxid, message.mxid)

    async def update_admin(self, update: UpdateChatParticipantAdmin) -> None:
        # TODO duplication not checked
        portal = await po.Portal.get_by_tgid(TelegramID(update.chat_id))
        if not portal or not portal.mxid:
            return

        await portal.set_telegram_admin(TelegramID(update.user_id))

    async def update_typing(
        self, update: UpdateUserTyping | UpdateChatUserTyping | UpdateChannelUserTyping
    ) -> None:
        sender = None
        if isinstance(update, UpdateUserTyping):
            portal = await po.Portal.get_by_tgid(
                TelegramID(update.user_id), tg_receiver=self.tgid, peer_type="user"
            )
            sender = await pu.Puppet.get_by_tgid(TelegramID(update.user_id))
        elif isinstance(update, UpdateChannelUserTyping):
            portal = await po.Portal.get_by_tgid(TelegramID(update.channel_id))
        elif isinstance(update, UpdateChatUserTyping):
            portal = await po.Portal.get_by_tgid(TelegramID(update.chat_id))
        else:
            return

        if isinstance(update, (UpdateChannelUserTyping, UpdateChatUserTyping)):
            sender = await pu.Puppet.get_by_peer(update.from_id)

        if not sender or not portal or not portal.mxid:
            return

        await portal.handle_telegram_typing(sender, update)

    async def _handle_entity_updates(self, entities: dict[int, User | Chat | Channel]) -> None:
        try:
            users = (entity for entity in entities.values() if isinstance(entity, (User, Channel)))
            puppets = ((await pu.Puppet.get_by_peer(user), user) for user in users)
            await asyncio.gather(
                *[puppet.try_update_info(self, info) async for puppet, info in puppets if puppet]
            )
        except Exception:
            self.log.exception("Failed to handle entity updates")

    async def update_others_info(self, update: UpdateUserName | UpdateUser) -> None:
        # TODO duplication not checked
        puppet = await pu.Puppet.get_by_tgid(TelegramID(update.user_id))
        if isinstance(update, UpdateUserName):
            if len(update.usernames) > 1:
                self.log.warning(
                    "Got update with multiple usernames (%s) for %s, only saving first one",
                    update.usernames,
                    update.user_id,
                )
            puppet.username = update.usernames[0].username if update.usernames else None
            if await puppet.update_displayname(self, update):
                await puppet.save()
                await puppet.update_portals_meta()
        elif isinstance(update, UpdateUser):
            info = await self.client.get_entity(puppet.peer)
            await puppet.update_info(self, info)
        else:
            self.log.warning(f"Unexpected other user info update: {type(update)}")

    async def update_status(self, update: UpdateUserStatus) -> None:
        puppet = await pu.Puppet.get_by_tgid(TelegramID(update.user_id))
        if isinstance(update.status, UserStatusOnline):
            await puppet.default_mxid_intent.set_presence(PresenceState.ONLINE)
        elif isinstance(update.status, UserStatusOffline):
            await puppet.default_mxid_intent.set_presence(PresenceState.OFFLINE)
        else:
            self.log.warning(f"Unexpected user status update: type({update})")
        return

    async def get_message_details(
        self, update: UpdateMessage
    ) -> tuple[UpdateMessageContent, pu.Puppet | None, po.Portal | None]:
        if isinstance(update, UpdateShortChatMessage):
            portal = await po.Portal.get_by_tgid(TelegramID(update.chat_id), peer_type="chat")
            sender = await pu.Puppet.get_by_tgid(TelegramID(update.from_id))
        elif isinstance(update, UpdateShortMessage):
            portal = await po.Portal.get_by_tgid(
                TelegramID(update.user_id), tg_receiver=self.tgid, peer_type="user"
            )
            sender = await pu.Puppet.get_by_tgid(self.tgid if update.out else update.user_id)
        elif isinstance(
            update,
            (
                UpdateNewMessage,
                UpdateNewChannelMessage,
                UpdateEditMessage,
                UpdateEditChannelMessage,
            ),
        ):
            update = update.message
            if isinstance(update, MessageEmpty):
                return update, None, None
            portal = await po.Portal.get_by_entity(update.peer_id, tg_receiver=self.tgid)
            if update.out:
                sender = await pu.Puppet.get_by_tgid(self.tgid)
            elif isinstance(update.from_id, (PeerUser, PeerChannel)):
                sender = await pu.Puppet.get_by_peer(update.from_id)
            elif isinstance(update.peer_id, PeerUser):
                sender = await pu.Puppet.get_by_peer(update.peer_id)
            else:
                sender = None
        else:
            self.log.warning(
                f"Unexpected message type in User#get_message_details: {type(update)}"
            )
            return update, None, None
        return update, sender, portal

    @staticmethod
    async def _try_redact(message: DBMessage) -> None:
        portal = await po.Portal.get_by_mxid(message.mx_room)
        if not portal:
            return
        try:
            await portal.main_intent.redact(message.mx_room, message.mxid)
        except MatrixError:
            pass

    async def delete_message(self, update: UpdateDeleteMessages) -> None:
        if len(update.messages) > self.max_deletions:
            return

        for message_id in update.messages:
            for message in await DBMessage.get_all_by_tgid(TelegramID(message_id), self.tgid):
                if message.redacted:
                    continue
                await message.delete()
                number_left = await DBMessage.count_spaces_by_mxid(message.mxid, message.mx_room)
                if number_left == 0:
                    await self._try_redact(message)

    async def delete_channel_message(self, update: UpdateDeleteChannelMessages) -> None:
        if len(update.messages) > self.max_deletions:
            return

        channel_id = TelegramID(update.channel_id)

        for message_id in update.messages:
            for message in await DBMessage.get_all_by_tgid(TelegramID(message_id), channel_id):
                if message.redacted:
                    continue
                await message.delete()
                await self._try_redact(message)

    async def update_reactions(self, update: UpdateMessageReactions) -> None:
        portal = await po.Portal.get_by_entity(update.peer, tg_receiver=self.tgid)
        if not portal or not portal.mxid or not portal.allow_bridging:
            return
        await portal.handle_telegram_reactions(self, TelegramID(update.msg_id), update.reactions)

    async def update_phone_call(self, update: UpdatePhoneCall) -> None:
        self.log.debug("Phone call update %s", update)
        if not isinstance(update.phone_call, PhoneCallRequested):
            return
        tgid = TelegramID(update.phone_call.participant_id)
        if tgid == self.tgid:
            tgid = update.phone_call.admin_id
        portal = await po.Portal.get_by_tgid(tgid, tg_receiver=self.tgid, peer_type="user")
        if not portal or not portal.mxid or not portal.allow_bridging:
            return
        sender = await pu.Puppet.get_by_tgid(TelegramID(update.phone_call.admin_id))
        await portal.handle_telegram_direct_call(self, sender, update)

    async def update_channel(self, update: UpdateChannel) -> None:
        portal = await po.Portal.get_by_tgid(TelegramID(update.channel_id))
        if not portal:
            return
        if getattr(update, "mau_telethon_is_leave", False):
            self.log.debug("UpdateChannel has mau_telethon_is_leave, leaving portal")
            await portal.delete_telegram_user(self.tgid, sender=None)
        elif chan := getattr(update, "mau_channel", None):
            if not portal.mxid:
                if (
                    not self.is_relaybot
                    or not self.config["bridge.relaybot.ignore_unbridged_group_chat"]
                ):
                    background_task.create(self._delayed_create_channel(chan))
            else:
                self.log.debug("Updating channel info with data fetched by Telethon")
                await portal.update_info(self, chan)
                await portal.invite_to_matrix(self.mxid)

    async def _delayed_create_channel(self, chan: Channel) -> None:
        self.log.debug(
            f"Waiting 5 seconds before handling UpdateChannel for non-existent portal {chan.id}"
        )
        await asyncio.sleep(5)
        portal = await po.Portal.get_by_tgid(TelegramID(chan.id))
        if portal.mxid:
            self.log.debug(
                "Portal started existing after waiting 5 seconds, "
                f"dropping UpdateChannel for {portal.tgid}"
            )
            return
        else:
            self.log.info(
                f"Creating Matrix room for {portal.tgid}"
                " with data fetched by Telethon due to UpdateChannel"
            )
            await portal.create_matrix_room(self, chan, invites=[self.mxid])

    async def _check_server_notice_edit(self, message: Message) -> None:
        pass

    async def update_message(self, original_update: UpdateMessage) -> None:
        update, sender, portal = await self.get_message_details(original_update)
        if not portal:
            return
        elif portal and not portal.allow_bridging:
            self.log.debug(
                f"Ignoring message {update.id} in portal {portal.tgid_log} (bridging disallowed)"
            )
            return

        if not portal.mxid and getattr(original_update, "mau_left_channel", False):
            self.log.debug(
                f"Ignoring message {update.id} in portal {portal.tgid_log} because user isn't in the chat"
            )
            return

        if self.is_relaybot:
            if update.is_private:
                if not self.config["bridge.relaybot.private_chat.invite"]:
                    if sender:
                        self.log.debug(f"Ignoring private message to bot from {sender.id}")
                    return
            elif not portal.mxid and self.config["bridge.relaybot.ignore_unbridged_group_chat"]:
                self.log.debug(
                    f"Ignoring message received by bot in unbridged chat {portal.tgid_log}"
                )
                return

        if (
            self.ignore_incoming_bot_events
            and self.relaybot
            and sender
            and sender.id == self.relaybot.tgid
        ):
            self.log.debug("Ignoring relaybot-sent message %s to %s", update.id, portal.tgid_log)
            return

        task = self._call_portal_message_handler(update, original_update, portal, sender)
        if portal.backfill_lock.locked:
            self.log.debug(
                f"{portal.tgid_log} is backfill locked, moving incoming message to async task"
            )
            background_task.create(task)
        else:
            await task

    async def _call_portal_message_handler(
        self,
        update: UpdateMessageContent,
        original_update: UpdateMessage,
        portal: po.Portal,
        sender: pu.Puppet,
    ) -> None:
        await portal.backfill_lock.wait(f"update {update.id}")

        if isinstance(update, MessageService):
            if isinstance(update.action, MessageActionChannelMigrateFrom):
                self.log.debug(
                    "Received %s in %s by %d, unregistering portal...",
                    update.action,
                    portal.tgid_log,
                    sender.id,
                )
                await self.unregister_portal(update.action.chat_id, update.action.chat_id)
                await self.register_portal(portal)
                return
            self.log.debug(
                "Handling action %s to %s by %d",
                update.action,
                portal.tgid_log,
                (sender.id if sender else 0),
            )
            return await portal.handle_telegram_action(self, sender, update)

        if isinstance(original_update, (UpdateEditMessage, UpdateEditChannelMessage)):
            if sender and sender.tgid == 777000:
                await self._check_server_notice_edit(update)
            return await portal.handle_telegram_edit(self, sender, update)
        return await portal.handle_telegram_message(self, sender, update)

    # endregion
